from __future__ import annotations
from ..tree import Tree

from ..constants import KEYCHAIN_LEFT_BOUND, KEYCHAIN_RIGHT_BOUND
from ..transformers import ValueTransformer

class AbstractValidator(Tree):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
    def invert(self):
        """Creates an inverted index of the configuration tree.

        Much like a textbook index, this method scans the tree for all variable
        tokens (e.g., `${VAR}` or `))VAR`) and generates a new tree that maps
        each variable to a list of all the locations (keychains) where it is
        used. This is highly useful for dependency analysis and debugging
        complex configurations.

        Returns:
            AbstractValidator: A new `AbstractValidator` instance representing the inverted index.
                The keys of this new tree are the variable names found in the
                original tree, and the values are lists of keychain strings
                indicating every location where each variable was used.

        Examples:
            >>> yaml_content = '''
            ... server:
            ...   host: ))host-ip
            ... database:
            ...   url: 'postgres://))db-user@))host-ip/))db-name'
            ... '''
            >>> yt = YAMLator(yaml_content) # Assuming YAMLator inherits from AbstractValidator
            >>> inverted_tree = yt.invert()
            >>> print(inverted_tree.pretty())
            ))host-ip:
            - server/host
            - database/url
            ))db-user:
            - database/url
            ))db-name:
            - database/url
        """
        _var_tree = self.__class__() # Use cls() for generic instantiation
        def _val(value, keychain):
            if not isinstance(value, str) and not isinstance(value, list):
                return
            if keychain and keychain[-1].startswith('_'):
                return
            if not isinstance(value, list):
                value = [value]
            for _value in value:
                _tokens = ValueTransformer()._tokenize(_value)
                for _token in _tokens:
                    _token,_keychain = self._remove_data_index(_token)
                    # _keychain == None means _token looks like a variable but is not a value type variable
                    if not _keychain: continue
                    # we cannot use .get() here because it will parse )){a/b} like a keychain string!
                    _tree_values = _var_tree.odict.get(_token, [])
                    _tree_values.append('/'.join(keychain))
                    _var_tree.odict[_token] = _tree_values

        self.visit(value_process=_val)

        return _var_tree

    def reduce(self):
        """Builds a dependency graph of the configuration.

        This method acts as a tool for creating a dependency graph, much like a
        "Bill of Materials" for your configuration. For each keychain, it
        determines which variables its value depends on.

        Its primary use case is to enable the detection of circular dependencies
        before running the full, potentially expensive, transformation process.
        Self-referential dependencies (where a key's value refers to itself)
        are ignored.

        Returns:
            AbstractValidator: A new `AbstractValidator` instance representing the dependency graph.
                It preserves the keychains from the original tree, but replaces
                their values with a list of the variable tokens they depend on.
                Keychains with no dependencies are omitted from the result.

        Examples:
            >>> yaml_content = '''
            ... db-host: 'db.internal'
            ... db-port: 5432
            ... db-url: 'postgres://user@))db-host:))db-port'
            ... app-port: ))db-port
            ... '''
            >>> yt = YAMLator(yaml_content) # Assuming YAMLator inherits from AbstractValidator
            >>> dependency_graph = yt.reduce()
            >>> print(dependency_graph.pretty())
            db-url:
            - '))db-host'
            - '))db-port'
            app-port:
            - '))db-port'
            <BLANKLINE>
        """
        _reduced_tree = self.__class__() # Use cls() for generic instantiation
        def _val(value, keychain):
            if not isinstance(value, str) and not isinstance(value, list):
                return
            if keychain and keychain[-1].startswith('_'):
                return
            if not isinstance(value, list):
                value = [value]

            _variables_in_value = []
            for _value in value:
                _tokens = ValueTransformer()._tokenize(_value)
                for _token in _tokens:
                    # this is bad: we want to parse ))a//b ( ))a/, /b )
                    if _token[-1] == '/':
                        _token = _token[:-1]
                    _token,_keychain = self._remove_data_index(_token)
                    # _keychain == None means _token looks like a variable but is not a value type variable
                    if not _keychain:
                        continue
                    if not _keychain == keychain[-1]:
                        _variables_in_value.append(_token)

            if _variables_in_value:
                _reduced_tree.get(keychain, _variables_in_value)

        self.visit(value_process=_val)

        return _reduced_tree

    def _remove_data_index(self,token):
        _token = token
        _keychain = None
        # there are variables like )){@[-1]/c} that look like variables but do not tokenize
        _m = ValueTransformer()._extract(token)
        if _m:
            _keychain, _data_index = _m.groups()
            if '/' in _keychain:
                _token =  KEYCHAIN_LEFT_BOUND + _keychain + KEYCHAIN_RIGHT_BOUND
                _keychain = _keychain.split('/')[-1]
            else:
                _token = _keychain
            _token = '))' + _token
        return _token,_keychain



from ..tree import *

class AbstractEvaluator(Tree):
    def __init__(self, odict_or_tree):
        super(AbstractEvaluator,self).__init__(odict_or_tree)

    def evaluate(self,*args):
        self.visit(self._pre_evaluate, self._post_evaluate, self._value_evaluate)

    def _pre_evaluate(self, node, keychain):
        pass

    def _post_evaluate(self, node, keychain):
        pass

    def _value_evaluate(self, value, keychain):
        pass



from ..constants import *
from .PathValueTransformer import *
from ..tree import *

from . import DEBUG

class YAMLTransformerException(Exception):
    pass


class YAMLTransformer(PathValueTransformer):
    name = 'transform_yaml'
    ext = 'yaml'

    def __init__(self, odict_or_yamlator_or_tree, context_tree=None, allow_tree_subs=False, root_dir=None):

        if hasattr(odict_or_yamlator_or_tree, 'root_dir') and root_dir is None:
            self.root_dir = odict_or_yamlator_or_tree.root_dir
        else:
            root_dir = pathlib.Path('.') if root_dir is None else root_dir
            self.root_dir = pathlib.Path(root_dir).absolute()

        super(YAMLTransformer, self).__init__(odict_or_yamlator_or_tree)

    def select(self, stream, keychain):

        if any(map(is_variable_token,keychain.split('/'))):
            if DEBUG.YAMLTransformer:
                _msg = f'found unsubbed vars'
                ic(_msg)
                ic(keychain)
            return None

        _stream_path = pathlib.Path(stream.name)
        if DEBUG.YAMLTransformer:
            ic()
            ic(keychain)
            ic(_stream_path)
        if _stream_path.absolute():
            _new_root_dir = _stream_path.parent
            _yt = YAMLTransformer(Tree.load(stream),root_dir=_new_root_dir)
        else:
            _new_root_dir = _stream_path.parent.relative_to(self.root_dir)
            _yt = YAMLTransformer(Tree.load(stream),self.root_dir.joinpath(_new_root_dir))

        if DEBUG.YAMLTransformer:
            ic(_new_root_dir)

        # TODO: this needs documenting; it's tricky
        # NOTICE: keeps loading trees until we run out of selector references
        # YAMLTransformer calls itself recursively

        _yt.evaluate()

        try:
            # if DEBUG.YAMLTransformer:
            #     _value = _yt.get(keychain)
                # ic(_value.print()) if isinstance(_value,Tree) else ic(_value)
            return _yt.get(keychain)
        except KeyError:
            # keychain may contain un-subbed variables
            if DEBUG.YAMLTransformer:
                _msg = f'Failed to select {keychain}'
                ic(_msg)
            return


class YAMLTransformerUtility(YAMLTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


from ..tree import *

class AbstractState(Tree):
    def __init__(self,  pre_observables=(), value_observables=(), post_observables=(),initial_state=None):
        super().__init__(initial_state)
        self.pre_observables = pre_observables
        self.post_observables = post_observables
        self.value_observables = value_observables
        self.all_observables = pre_observables + post_observables + value_observables

    def pre_update(self,node,keychain):
        raise NotImplemented

    def value_update(self,value,keychain):
        raise NotImplemented

    def post_update(self,node,keychain):
        raise NotImplemented


import io
import re
import ast
import yarl
import yaml
import astor
import pathlib
import typing

from copy import copy,deepcopy
from collections import OrderedDict
from functools import partial,reduce
from dataclasses import dataclass
from enum import Enum


try:
    from icecream import ic
    ic.configureOutput(includeContext=True)
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

KEYCHAIN_SEP = r'/'
KEY_OR_KEYCHAIN_OP = r'))'
KEYCHAIN_LEFT_BOUND = r'{'
KEYCHAIN_RIGHT_BOUND = r'}'

TRUE_VALUES = \
    ("y","Y","yes","Yes","YES",)\
    + ("true","True","TRUE",)\
    + ("on","On","ON",)

FALSE_VALUES = \
    ("n","N","no","No","NO",)\
    + ("false","False","FALSE",)\
    + ("off","Off","OFF")


class REGEXES:
    BOOL_VALUE = r'|'.join(TRUE_VALUES) + r'|' + r'|'.join(FALSE_VALUES)

    KEY = r'[a-zA-Z0-9_][a-zA-Z0-9_\.-]*'
    KEYCHAIN = rf'/?(?:{KEY}/)*'

    BOOL_TYPE_KEY = rf'(?:use|is)-{KEY}'
    PATH_TYPE_KEY = rf'{KEY}-(?:path|dir)$'
    URL_TYPE_KEY = rf'{KEY}-url$'

    LIST_INDEX = r'\d+'
    SLICE = r'\:?\-?\d+\:?'
    DATA_INDEX = rf'{KEYCHAIN}{KEY}/?|{SLICE}'
    POSIX_RELATIVE = r'\./(?:[^\0/]+/?)*[^#]'
    POSIX_ABSOLUTE = r'/(?:[^\0/]+/?)*[^#]'

    AT = r'@(?:\[\-?\d+\])?'
    AT_EXP = r'@(?:\[(\-?)(\d+)\])?'

    # FRAGILE! Does not allow multiple bang expressions on one line
    BANG = r'(?:\!.*\))'
    BANG_EXP =r'\!([^)]*?\)$)'

    IMPORT = r'(?:\+.*\)(?:\s|$))'
    IMPORT_EXP = r'\+.*(?:\s|$)'
    # -----------------------------------

    IF = rf'\?{re.escape(KEYCHAIN_LEFT_BOUND)}[^}}]+{re.escape(KEYCHAIN_RIGHT_BOUND)}'

    OR_SEP = r'|'
    AND_SEP = r'&'
    CASE_SEP = r':'

    LOGICAL_KEY_EXP = \
        rf'[{OR_SEP}{AND_SEP}]?\s*!?\s*[^=!{AND_SEP}{OR_SEP}]+\s*(?:[!=]=\s*!?\s*[^=!{AND_SEP}{OR_SEP}]+)?'

    IF_KEY_EXP = \
        rf'\?{re.escape(KEYCHAIN_LEFT_BOUND)}((?:\s*{LOGICAL_KEY_EXP}|\s*[{OR_SEP}{AND_SEP}]?\s*!?\s*{BOOL_TYPE_KEY})+)\s*{re.escape(KEYCHAIN_RIGHT_BOUND)}(/?)'

    LOGICAL_EXP = \
        rf'[{OR_SEP}{AND_SEP}]?\s*!?\s*[^=!{AND_SEP}{OR_SEP}{CASE_SEP}]+(?:\[{DATA_INDEX}\])?\s*(?:[!=]=\s*!?\s*[^=!{AND_SEP}{OR_SEP}{CASE_SEP}]+)?'

    IF_EXP = \
        rf'\?{re.escape(KEYCHAIN_LEFT_BOUND)}((?:\s*{LOGICAL_EXP}|\s*[{OR_SEP}{AND_SEP}]?\s*!?\s*{BOOL_TYPE_KEY})+)\s*{CASE_SEP}' + \
        rf'([^{AND_SEP}{OR_SEP}{CASE_SEP}]+)\s*(?:{CASE_SEP}([^{AND_SEP}{OR_SEP}{CASE_SEP}]+))?\s*{re.escape(KEYCHAIN_RIGHT_BOUND)}'


def bool_factory(x):
    if not re.match(REGEXES.BOOL_VALUE, x):
        raise f"{x} is not a recognized boolean string"
    if x in TRUE_VALUES:
        x = True
    elif x in FALSE_VALUES:
        x = False
    return x


DEFAULT_KEY_TYPES = dict(
    path = (REGEXES.PATH_TYPE_KEY, (pathlib.Path,)),
    url = (REGEXES.URL_TYPE_KEY,(yarl.URL,)),
    bool = (REGEXES.BOOL_TYPE_KEY,(bool_factory,))
)


def is_variable_token(token):
    if str(token)[:len(KEY_OR_KEYCHAIN_OP)] == KEY_OR_KEYCHAIN_OP:
        return True
    return False



from ..constants import *
from .AbstractObservable import *

from . import DEBUG

class KeyPresenceObservable(AbstractObservable):
    keys = None

    def _test_keychain(self,node,keychain):
        if any(map(lambda x:x in keychain,self.keys)):
            return keychain
        return


class KeyAbsenceObservable(AbstractObservable):
    keys = None

    def _test_keychain(self, node,keychain):
        if all(map(lambda x: not x in keychain, self.keys)):
            return keychain
        return


class KeyLookAheadObservable(AbstractObservable):
    keys = None

    def _test_keychain(self,node,keychain):
        if DEBUG.Observable:
            ic(self.__class__)
            ic(keychain)

        _matched_keychain = None
        _matched_node_key = None
        for _node_key in node.keys():

            if _node_key in self.keys:

                assert _matched_node_key is None
                _matched_node_key = _node_key

        if _matched_node_key:
            _matched_keychain = keychain + [_matched_node_key]

        if DEBUG.Observable:
            ic(_matched_node_key)
            ic(_matched_keychain)
        return _matched_keychain


class RegexObservable(AbstractObservable):
    keychain_regex = None

    def _test_keychain(self, node, keychain):
        if DEBUG.Observable:
            ic(self.__class__)
            ic(keychain)
        _m = re.match(self.keychain_regex, '/'.join(keychain))
        if not _m:
            return
        _parameters = _m.groups()
        return _parameters


class RegexLookAheadObservable(AbstractObservable):
    keychain_regex = None

    def _test_keychain(self,node, keychain):
        _parameters = None
        _matched_keychain = None
        _matched_node_key = None
        for _node_key in node.keys():
            _m = re.match(self.keychain_regex,'/'.join(keychain + [_node_key]))
            if _m:
                _parameters = _m.groups()

        return _parameters




from ..constants import *
from ..evaluators.AbstractEvaluator import *

from . import DEBUG

class TransformerException(Exception):
    pass

class Transformer(AbstractEvaluator):

    def __init__(self, odict_or_tree=None):
        odict_or_tree = OrderedDict() if not odict_or_tree else odict_or_tree
        # super(Transformer, self).__init__(odict_or_tree)
        super().__init__(odict_or_tree)
    def _match(self, line):
        '''used recursively in self._tokenize to split a line into tokens by matching'''
        # return a re.match object or None to skip
        # if m = re.match then  len(m.groups()) == 3: head,match,tail
        raise NotImplemented

    def _extract(self, token):
        '''used in self.value_evaluate() to extract regex groups from a token which are passed to self._transform() as parameters'''
        # return a re.match object or None to skip
        raise NotImplemented

    def _transform(self, parameters, keychain):
        # use the supplied parameters to produce a str or bytes value
        # include the keychain for context awareness
        raise NotImplemented

    def _do_not_evaluate(self, value, keychain):
        # skip this nodes for a specific reason
        return False

    def _tokenize(self, expression, pieces=None):
        '''
        tokenize an expression recursively, splitting it by presence of the first match of re_matcher expression
        turn a line into a list of tokens
        '''
        if DEBUG.Transformer:
            ic()
            ic(self.match_regex)
            ic(expression)
            # ic(pieces)
        # don't parse blank lines; but include them in the output
        if expression == '':
            return ''

        pieces = pieces if pieces is not None else []

        _match = self._match(expression)
        if _match:
            if DEBUG.Transformer:
                ic(_match.groups())

            _front, _match, _back = _match.groups()
            if _match == '':
                pieces.append(expression)
            if _front != '':
                pieces.append(_front)

            pieces.append(_match)
            if _back != '':
                self._tokenize(_back, pieces)
        else:
            pieces.append(expression)

        if DEBUG.Transformer:
            ic(pieces)

        return list(pieces)



class AbstractObservable:
    # we might want to turn off certain observables from aggregating
    aggregate = True

    def __init__(self,**kwargs):
        [ setattr(self,k,v) for k,v in kwargs.items() ]

    def observe(self,node,keychain):
        _keychain_parameters = self._test_keychain(node,keychain)
        if _keychain_parameters:
            return self._create_observation(node,*_keychain_parameters)
        return None

    def read(self,state_or_evaluator):
       pass

    def _test_keychain(self,node,keychain):
        return []

    def _create_state_keychain(self,*parameters):
        return

    def _create_observation(self,node,*parameters):
        return self._create_state_keychain(*parameters),None




from .constants import *

class _DEBUG:
    VISIT = False
    POP = False
    GET = False
    FIND = False
    FLATTEN = False
    STRINGIFY = False
    OVERLAY = False
    DFS = False
    INDENT = None
    RESET = False
    ENTRY = False


class TreeVisitRestartException(Exception):
    pass


class TreeVisitStopException(Exception):
    pass


class TreeException(Exception):
    pass



class TreeCreationException(TreeException):
    pass


class Tree:
    def __eq__(self, other):
        if self.odict == other.odict:
            return True
        return False

    def __add__(self, other):
        self.overlay(other)
        return self

    def __str__(self):
        #this has to be dump(); other functions depend upon this string rep
        # not the same as self.print() !
        return self.dump()

    def __init__(self,odict_or_tree=None,**kwargs):
        if odict_or_tree is None:
            self.odict = OrderedDict()
            # TODO: what is the point of this?
            #-------------------------------
            if kwargs:
                for _k,_v in kwargs.items():
                    self.get(_k,_v)
            #-------------------------------

        elif isinstance(odict_or_tree,OrderedDict):
            self.odict = odict_or_tree
        elif issubclass(odict_or_tree.__class__,Tree):
            self.odict = odict_or_tree.odict
        elif isinstance(odict_or_tree,str):
            self.odict = self.load(io.StringIO(odict_or_tree)).odict
        elif isinstance(odict_or_tree,dict):
            self.odict = OrderedDict(odict_or_tree)
        else:
            raise TreeCreationException('Cannot create tree')
        if _DEBUG.GET:
            _DEBUG.INDENT = -1

    def is_empty(self):
        if not self.odict:
            return True

    def copy(self):
        """Copies the Tree object.

        Returns:
             Tree: A deepcopy of the Tree object.
        """
        return self.__class__(deepcopy(self.odict))

    def reset(self, *keys_to_preserve):
        """Clears all contents of the tree, with an option to preserve branches.

        The tree is modified in-place.

        Args:
            *keys_to_preserve (str): A variable number of keychain strings.
                If no arguments are provided, the entire tree is emptied.
                If one or more keychains are provided, only those branches
                will remain in the tree.

        Returns:
            None

        Examples:
            >>> tree = Tree({'a': 1, 'b': {'c': 2}})
            >>> tree.reset()
            >>> tree.is_empty()
            True

            >>> tree = Tree({'a': 1, 'b': {'c': 2}, 'd': 3})
            >>> tree.reset('b/')
            >>> print(tree.dump())
            b:
              c: 2
        """
        _old_tree = self.copy()

        if _DEBUG.RESET:
            ic()
            ic(keys_to_preserve)
            ic(_old_tree.print())

        # preserve refs!
        _keys = list(self.odict.keys())
        _keys.reverse()
        for _key in _keys:
            # clean out the current odict but preserve its reference
            self.odict.pop(_key)
            if _DEBUG.RESET:
                print(self.odict)

        for _key_to_preserve in keys_to_preserve:
            try:
                _dfs_keychain,_dfs_value = _old_tree.dfs(_key_to_preserve)
                if isinstance(_dfs_value,Tree) and _dfs_keychain[-1] != '':
                    _dfs_keychain.append('')
                if _DEBUG.RESET:
                    ic(_dfs_keychain)
                    ic(_dfs_value)
                self.get(_dfs_keychain,_old_tree.get(_dfs_keychain))
            except KeyError:
                if _DEBUG.RESET:
                    _msg = f'{_key_to_preserve} not found in state'
                    ic(_msg)
                pass

        if _DEBUG.RESET:
            ic(self.print())

    def visit(self,pre_process=lambda x, y:None, post_process=lambda x, y:None, value_process=lambda x, y:None,reverse=False,entry_keychain=None):
        """Traverses the tree using a visitor pattern.

        This method walks through the entire tree, executing callbacks at
        different stages of the traversal. It provides three distinct callback
        hooks:

        - `pre_process`: Called on a branch node before its children are visited
            (pre-order traversal).
        - `post_process`: Called on a branch node after its children have been
            visited (post-order traversal).
        - `value_process`: Called on a leaf node.

        Args:
            pre_process (callable, optional): A function that accepts a branch
                node (OrderedDict) and its keychain (list[str]). It is
                executed before visiting the node's children. Defaults to a
                no-op.
            post_process (callable, optional): A function with the same
                signature as `pre_process`, but executed after visiting the
                node's children. Defaults to a no-op.
            value_process (callable, optional): A function that accepts a leaf
                value (any type) and its keychain (list[str]). Defaults to a
                no-op.
            reverse (bool, optional): If True, the children of each node are
                visited in reverse order. Defaults to False.
            entry_keychain (typing.Union[str, list[str]], optional): A specific
                keychain (string or list) that defines a starting point for
                processing. The traversal still begins at the root, but the
                callbacks will only be activated for the target node and its
                descendants. Defaults to None.

        Raises:
            TreeVisitRestartException: A callback can raise this exception to
                interrupt and restart the entire visit, optionally providing a
                new entry keychain to begin the new traversal from.
            TreeVisitStopException: A callback can raise this exception to
                gracefully and immediately terminate the entire traversal.

        Examples:
            >>> tree = Tree({'a': {'b': 1}, 'c': 2})
            >>>
            >>> def pre(node, keychain):
            ...     indent = '  ' * len(keychain)
            ...     if keychain:
            ...         print(f"{indent}{keychain[-1]}:")
            ...
            >>> def post(node, keychain):
            ...     pass  # Not needed for this example
            ...
            >>> def value(val, keychain):
            ...     indent = '  ' * (len(keychain) -1)
            ...     print(f"{indent}  {keychain[-1]}: {val}")
            ...
            >>> tree.visit(pre_process=pre, value_process=value)
            a:
              b: 1
            c: 2
        """
        try:
            if not reverse:
                self._visit(self.odict,pre_process,post_process,value_process,None,entry_keychain=entry_keychain)
            else:
                self._reverse_visit(self.odict,pre_process,post_process,value_process,None)
        except TreeVisitRestartException as e:
            if _DEBUG.VISIT or _DEBUG.ENTRY:
                ic(e)
                ic(self.odict)
            if e.args[0] is not None:
                _keychain_or_keychain_str = e.args[0]
                _entry_keychain = copy(_keychain_or_keychain_str) if isinstance(_keychain_or_keychain_str,list) else _keychain_or_keychain_str.split('/')
                if _DEBUG.ENTRY:
                    ic(_entry_keychain)
                self.visit(pre_process, post_process, value_process, reverse,entry_keychain=_entry_keychain)
            else:
                self.visit(pre_process, post_process, value_process, reverse)
        except TreeVisitStopException as e:
            if _DEBUG.VISIT:
                ic(e)
            return

    def _visit(self, node, pre_process=lambda x, y:None, post_process=lambda x, y:None, value_process=lambda x, y:None,keychain=None,entry_keychain=None,process=False):
        keychain = [] if keychain is None else keychain
        entry_keychain = [] if entry_keychain is None else entry_keychain
        if not entry_keychain:
            process = True
        elif not process and keychain[:len(entry_keychain)] == entry_keychain:
            process = True
            if _DEBUG.ENTRY:
                _msg = f"{keychain[:len(entry_keychain)]} == {entry_keychain} "
                ic(_msg)
                ic(process)

        if isinstance(node,OrderedDict):
            if process:
                pre_process(node,copy(keychain))
            for _child_key in node.keys():
                keychain.append(_child_key)
                self._visit(node.get(_child_key), pre_process, post_process, value_process, copy(keychain),entry_keychain,process)
                keychain.pop()
            if process:
                post_process(node,copy(keychain))
        else:
            if process:
                value_process(node,keychain)

    def _reverse_visit(self, node,pre_process=lambda x, y:None, post_process=lambda x, y:None, value_process=lambda x, y:None,keychain=None):
        keychain = [] if keychain is None else keychain
        if isinstance(node,OrderedDict):
            pre_process(node,copy(keychain))
            _node_keys = copy(list(node.keys()))
            _node_keys.reverse()
            for _child_key in _node_keys:
                keychain.append(_child_key)
                self._reverse_visit(node.get(_child_key), pre_process, post_process, value_process, copy(keychain))
                keychain.pop()
            post_process(node,copy(keychain))
        else:
            value_process(node,keychain)

    def keys(self, keychain_or_keychain_str=None):
        """Returns the keys of a node, behaving like dict.keys().

        Args:
            keychain_or_keychain_str (typing.Union[str, list[str]], optional):
                A keychain string or list of strings specifying the target node.
                If omitted, returns the keys of the root node. Defaults to None.

        Returns:
            dict_keys: A view object containing the keys of the specified node.

        Raises:
            KeyError: If the keychain is invalid or points to a leaf node.

        Examples:
            >>> tree = Tree({'a': {'b': 'B_VAL'}, 'c': 'C_VAL'})
            >>> list(tree.keys())
            ['a', 'c']
            >>> list(tree.keys('a/'))
            ['b']
        """
        if keychain_or_keychain_str:
            _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str,list) else keychain_or_keychain_str.split('/')
            # never do this
            # if not _keychain[-1] == '':
            if _keychain[-1] != '':
                _keychain += ['']
        try:
            return self.odict.keys() if keychain_or_keychain_str is None else self.get(_keychain).keys()
        except AttributeError:
            raise KeyError

    def dfs(self, keychain_or_keychain_str, keychain_or_keychain_str_context=None):
        """Performs a depth-first search (DFS) to locate a node in the tree.

        The search is performed for the first key in the provided keychain. If
        multiple nodes with the same key exist, the search prioritizes and
        returns the one that is nested deepest within the tree structure.

        Args:
            keychain_or_keychain_str (typing.Union[str, list[str]]): The keychain
                to search for.
            keychain_or_keychain_str_context (typing.Union[str, list[str]], optional):
                An optional keychain that defines a starting branch for the search.
                If provided, the search is restricted to descendants of this node.
                Defaults to None.

        Returns:
            tuple[list[str], typing.Any]: A tuple containing the full, absolute
            keychain (as a list of strings) of the found node and the value at
            that node.

        Raises:
            KeyError: If the keychain cannot be found in the tree.
        """
        _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str,list) \
                                                                else keychain_or_keychain_str.split('/')

        # if DEBUG.DFS:
        #     ic(keychain_or_keychain_str)
        #     ic(keychain_or_keychain_str_context)

        keychain_or_keychain_str_context = [] \
            if keychain_or_keychain_str_context is None else keychain_or_keychain_str_context

        # if DEBUG.DFS:
        #     ic(keychain_or_keychain_str)
        #     ic(keychain_or_keychain_str_context)

        _keychain_context = copy(keychain_or_keychain_str_context) \
            if isinstance(keychain_or_keychain_str_context,list) else keychain_or_keychain_str_context.split('/')

        if _DEBUG.DFS:
            ic(_keychain)
            ic(_keychain_context)

        if not _keychain or (len(_keychain) == 1 and _keychain[0] == ''):
            # if _keychain[0] == '' :
            if _DEBUG.DFS:
                _msg = f'returning /,{self.get(_keychain_context)}'
                ic(_msg)
            return '/',self.get(_keychain_context)

        if _keychain[0] == '':
            _keychain = _keychain[1:]

        key = _keychain[0]

        _all_results = []

        def _dfs(node,keychain):
            if _DEBUG.DFS:
                ic()
                ic(keychain)

            if not keychain[:len(_keychain_context)] == _keychain_context:
                # we are not a child of the given keychain context
                if _DEBUG.DFS:
                    _msg = f'returning None. {keychain} is not a child of {_keychain_context}'
                    ic(_msg)
                return

            _node_keys = copy(list(node.keys()))
            _node_keys.reverse()
            for _node_key in _node_keys:
                if _DEBUG.DFS:
                    ic(key)
                    ic(_node_key)
                if _node_key == key:
                    if _DEBUG.DFS:
                        _msg = f'found {key}'
                        ic(_msg)
                    # all results are absolute!
                    _all_results.append(([''] + keychain+[_node_key],node[_node_key]))

        self.visit(post_process=_dfs)

        if not _all_results:
            if _DEBUG.DFS:
                ic()
                ic(KeyError)
            raise KeyError

        # if DEBUG.DFS:
        #     ic(_all_results)

        _all_results.sort(key=lambda x:len(x[0]),reverse=True)

        if _DEBUG.DFS:
            ic(_all_results)

        if len(_keychain) > 1:
            # we are looking for a particular keychain
            for _result in _all_results:
                try:
                    # test for existence of keychain in search result
                    _value = Tree(_result[1]).get(_keychain[1:])
                    if _DEBUG.DFS:
                        _msg = f'returning {_result[0] + _keychain[1:]},{_value}'
                        ic(_msg)
                    return _result[0] + _keychain[1:],_value
                except KeyError:
                    pass
                if _DEBUG.DFS:
                    ic()
                    ic(KeyError)

                raise KeyError
        else:
            if _DEBUG.DFS:
                _msg = f'returning {_all_results[0]}'
                ic(_msg)
            return _all_results[0]

    def uget(self, keychain_or_keychain_str, value=None):
        """Gets a value from the tree, returning a default if the key does not exist.

        This method acts as a "get with default." It attempts to retrieve the
        value at the specified keychain. If the keychain does not exist, it
        sets the provided `value` at that keychain and then returns it.

        Args:
            keychain_or_keychain_str (typing.Union[str, typing.List[str]]): The
                keychain (path) to retrieve.
            value (typing.Any, optional): The default value to set and return if
                the keychain does not exist. Defaults to None.

        Returns:
            typing.Any: The value at the keychain if it exists, otherwise the
            provided default `value`.
        """
        try:
            #  get the value if it exists, forget the passed in value
            _value = self.get(keychain_or_keychain_str)
        except KeyError:
            #  create a key and set the value
            _value = self.get(keychain_or_keychain_str,value)
        return _value

    def aget(self,keychain_or_keychain_str, value):
        """Appends a value to a list at a specified keychain.

        If the keychain does not exist, a new list is created with `value` as
        its first element. If a value at the keychain exists but is not a
        list, it is converted into a list, and the new `value` is appended.
        If `value` itself is a list, its items are added to the list at the
        keychain.

        Args:
            keychain_or_keychain_str (typing.Union[str, typing.List[str]]): The
                keychain (path) where the value will be appended.
            value (typing.Any): The value or list of values to append.

        Returns:
            list: The modified list at the specified keychain.
        """
        if not keychain_or_keychain_str:
            # does not work at root level
            return
        if not isinstance(value,list):
            value = [value]

        # always trim the root on aget, but we call self.get on not trimmed root
        _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str, list) \
            else keychain_or_keychain_str.split('/')
        if not _keychain[-1] == '':
            _keychain.append('')
        else:
            keychain_or_keychain_str = keychain_or_keychain_str[:-1]

        try:
            _value = self.get(_keychain)
            if not isinstance(_value,list):
                _value = [_value]
            _value += value
        except KeyError:
            _value = value

        # call get() on the root of the keychain
        self.get(keychain_or_keychain_str,_value)
        # but return the value
        return _value

    def daget(self,keychain_or_keychain_str, value=None):
        """Updates a dictionary at a specified keychain.

        This method performs a dictionary update at the given keychain. If the
        keychain does not exist, a new dictionary is created. If a value
        exists at the keychain, it must be a dictionary, or a `TreeException`
        will be raised.

        Args:
            keychain_or_keychain_str (typing.Union[str, typing.List[str]]): The
                keychain (path) of the dictionary to update.
            value (dict): A dictionary of key-value pairs to update the target
                dictionary with.

        Returns:
            Tree: A `Tree` object representing the updated dictionary.

        Raises:
            TreeException: If the keychain is empty or if the existing value at
                the keychain is not a dictionary.
        """
        if not keychain_or_keychain_str:
            # does not work at root level
            raise TreeException('bad keychain!')
            # return
        assert isinstance(value,dict)
        try:
            _value = self.get(keychain_or_keychain_str)
            if not isinstance(_value, dict):
                # _value = [_value]
                raise TreeException('bad value!')
            _value.update(value)
        except KeyError:
            _value = value  # like a regular get() wth the key just created

        return self.get(keychain_or_keychain_str, _value)

    def get(self, keychain_or_keychain_str, value=None):
        """Gets or sets a value in the tree, with path-aware logic.

        This method serves a dual purpose:
        1.  **Getter (value is None):** Retrieves a node or subtree from the
            tree based on the provided keychain.
        2.  **Upsert Setter (value is not None):** Inserts or updates a value
            at the specified keychain. If the path does not exist, it will be
            created.

        Path Resolution:
        -   **Absolute Paths:** If `keychain_or_keychain_str` starts with a `/`,
            the path is resolved from the root of the tree.
        -   **Relative Paths:** If the path does not start with a `/`, the
            method performs a depth-first search (DFS) to find the first
            node matching the initial key, and then resolves the rest of the
            path from there.

        Trailing Slash Convention (for getters):
        -   **With trailing slash (e.g., 'a/b/'):** Returns the actual node
            at the given path. If the node is a branch, it returns a `Tree`
            object representing the subtree. If it's a leaf, it returns the
            leaf's value.
        -   **Without trailing slash (e.g., 'a/b'):** Returns a `Tree` object
            containing the target node as a value, with the last segment of
            the key as its key. For `x.get('a/b')`, the result is
            `Tree(OrderedDict(b=<value at keychain a/b in Tree x>))`.

        Args:
            keychain_or_keychain_str (typing.Union[str, typing.List[str]]): The keychain
                (path) to the target node. Can be a '/'-separated string or a
                list of strings.
            value (typing.Any, optional): The value to set at the keychain. If None,
                the method acts as a getter. Defaults to None.

        Returns:
            typing.Union['Tree', typing.Any]: When getting, returns a `Tree` object or a
            scalar value depending on the path and the trailing slash
            convention. When setting, it returns the value that was set.

        Raises:
            KeyError: If `value` is `None` and the specified keychain cannot be
                found in the tree.
            TreeException: If an attempt is made to set a scalar value at the
                root of the tree.

        Examples:
            >>> tree = Tree()
            >>> tree.get('/a/b', 'value1')
            >>> print(tree.get('/a/b/'))
            value1
            >>> print(tree.get('/a/b'))
            b:
              value1

            >>> tree.get('/a/c', {'d': 'value2'})
            >>> print(tree.dump())
            a:
              b: value1
              c:
                d: value2
        """
        # it should be passed in as a list to avoid splitting
        _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str,list) \
                                                                else keychain_or_keychain_str.split('/')
        _absolute_keychain = False
        if _keychain and _keychain[0] == '':
            _absolute_keychain = True
            _keychain = _keychain[1:]
        elif not _keychain:
            _absolute_keychain = True

        if _DEBUG.GET:
            ic()
            ic(keychain_or_keychain_str)
            ic(_absolute_keychain)
            ic(_keychain)
            ic(value)

        _trim_root = False
        # _keychain = [] if _keychain is None else _keychain

        if len(_keychain) > 1 and _keychain[-1] == '':
            # so if _keychain was None or [] we are
            _trim_root = True
            _keychain.pop()
        elif _keychain == ['']:
            _keychain.pop()

        if not _keychain and not isinstance(value,Tree) and value is not None:
            raise TreeException("Cannot set scalars at the root")

        _single_key = False

        if len(_keychain) == 1:
            _single_key = True

        _original_keychain = copy(_keychain)

        _keychain.reverse()

        if _DEBUG.GET:
            ic(_trim_root)
            ic(_original_keychain)
            ic(_keychain)
            if isinstance(value,Tree):
                ic(value.odict)
            else:
                ic(value)

        if _absolute_keychain:
            _ROOT_KEY = 'ROOT'

            _key = _ROOT_KEY
            _current_node_parent = self.odict
            _parents_key = _key # _key_parent[_key] == _key_parent ??? the root node?
            _current_node = _current_node_parent
        else:
            _ROOT_KEY = _keychain[0]
            _key = _ROOT_KEY
            _current_node_parent = None
            _parents_key = _key
            _current_node = _current_node_parent

        if not _keychain:
            if not _absolute_keychain:
                print('please use / to designate the root node')
                _absolute_keychain = True
            # we are manipulating the root node
            if value is not None:
                # it must be a Tree; not allowed to set scalars at root
                if _DEBUG.GET:
                    _msg = f'maninpulating the root node'
                    ic(_msg)
                    ic(_key)
                    ic(_parents_key)

                    ic(value.odict)

                _current_node.update({_k:_v for _k,_v in value.odict.items()})

                _current_node = value.odict


        else:
            if _absolute_keychain:
                while _keychain:
                    _key = _keychain.pop()
                    if _DEBUG.GET:
                        ic(_key)
                        ic(_keychain)
                        ic(_parents_key)
                        # ic(_current_node_parent)
                        # ic(_current_node)
                    try:
                        # _current_node.keys() will throw AttributeError if it is not an OrderedDict
                        if _key in _current_node.keys():
                            if _DEBUG.GET:
                                _msg = f'{_key} in _current_node keys'
                                ic(_msg)
                            # try to get the next element of the keychain
                            _current_node_parent = _current_node
                            _current_node = _current_node[_key]
                            _parents_key = _key
                        else:
                            if _DEBUG.GET:
                                _msg = f'{_key} not in _current_node keys'
                                ic(_msg)
                            raise KeyError
                    except (KeyError,AttributeError) as e:
                        if _DEBUG.GET:
                            ic(e)
                        if isinstance(e,AttributeError) and value:
                            if _DEBUG.GET:
                                _msg = f'setting _current_node_parent[{_parents_key}] = OrderedDict()'
                                ic(_msg)
                                _msg = f'setting _current_node = _current_node_parent[{_parents_key}]'
                                ic(_msg)

                            _current_node_parent[_parents_key] = OrderedDict()
                            _current_node = _current_node_parent[_parents_key]

                            _keychain.append(_key)
                            if _DEBUG.GET:
                                _msg = f'repeating key {_key}'
                                ic(_msg)

                            continue

                        elif isinstance(e,KeyError):
                            if _DEBUG.GET:
                                _msg = f'{_key} not found'
                                ic(_msg)

                            if value is not None:
                                if _DEBUG.GET:
                                    # _msg = f'Setting _current_node[{_key}] to None'
                                    _msg = f'Setting _current_node[{_key}] to OrderedDict()'
                                    ic(_msg)
                                    _msg = f'Setting _current_node_parent to _current_node'
                                    ic(_msg)
                                    _msg = f'Setting _current_node to _current_node[{_key}]'
                                    ic(_msg)

                                _current_node[_key] = OrderedDict()
                                _current_node_parent = _current_node
                                _current_node = _current_node[_key]
                                _parents_key = _key

                            elif value is None:
                                if _DEBUG.GET:
                                    _msg = f' raising KeyError'
                                    ic(_msg)
                                raise KeyError
                        else:
                            # raise e
                            # if we try to apply a key to a string value, it's really a KeyError, not AttributeError
                            raise KeyError

            else:
                _key  = _keychain.pop()

                try:
                    if _DEBUG.GET:
                        _msg = f'dfs looking for {_key}'
                        ic(_msg)
                    _dfs_keychain,_dfs_value = self.dfs(_key)

                    _relative_keychain = copy(_keychain)
                    _relative_keychain.reverse()
                    if _DEBUG.GET:
                        _msg = f'dfs found {_key} at:'
                        ic(_dfs_keychain)
                        # ic(_dfs_value)
                        ic(_relative_keychain)
                    #
                    if _trim_root:
                        _relative_keychain += ['']

                    if _DEBUG.GET:
                        _msg = f'RECURSING'
                        ic(_msg)

                    return self.get(_dfs_keychain+_relative_keychain,value)

                except KeyError:
                    # we could not find a relative root;
                    # assuming it is meant to be at root can be useful
                    if _DEBUG.GET:
                        _msg = f"Assuming {keychain_or_keychain_str} should have been absolute"
                        ic(_msg)
                    return self.get([''] + _original_keychain,value)

        if _DEBUG.GET:
            ic(_parents_key)
            # ic(_current_node_parent)
            # ic(_current_node)

        if value is not None and not isinstance(value,Tree):
            # if isinstance(_current_node,OrderedDict):
            if _DEBUG.GET:
                _msg = f'Setting _parent_current_node[{_key}] to {value}'
                ic(_msg)
                _msg = f'Setting _current_node = _parent_current_node[{_key}]'
                ic(_msg)
            _current_node_parent[_key] = value
            _current_node = _current_node_parent[_key]
            # else:
            #     if DEBUG.GET:
            #         _msg = f'Setting _parent_current_node[{_key}] to {value}'
            #         ic(_msg)
            #         _msg = f'Setting _current_node = _parent_current_node[{_key}]'
            #         ic(_msg)
            #     _current_node_parent[_key] = value
            #     _current_node = _current_node_parent[_key]
        elif value is not None and isinstance(value,Tree):
            if isinstance(_current_node,OrderedDict):
                if _DEBUG.GET:
                    _msg = f'updating _current_node with {list(value.keys())} from  value'
                    ic(_msg)
                _current_node.update({_k:_v for _k,_v in value.odict.items()})
            else:
                if _DEBUG.GET:
                    _msg = f'setting _current_node to OrderedDict()'
                    ic(_msg)
                    _msg = f'updating _current_node with {list(value.keys())} from value'
                    ic(_msg)
                    _msg = f'setting _current_node_parent[_key] = _current_node'
                _current_node = OrderedDict()
                _current_node.update({_k:_v for _k,_v in value.odict.items()})
                _current_node_parent[_key] = _current_node

        if not isinstance(_current_node,OrderedDict):
            if isinstance(_current_node,list) and not _trim_root:
                _ret = Tree(OrderedDict({_key: _current_node}))
            else:
                _ret = _current_node
        else:
            if not _trim_root:
                _ret = Tree(OrderedDict({_key:_current_node}) if _key != _ROOT_KEY else Tree(_current_node))
            else:
                _ret = Tree(_current_node)

        return _ret

    def pop(self, keychain_or_keychain_str, delete_key=True, keychain_context=None):
        """Removes and returns a node and its descendants from the tree.

        This method locates a node by its keychain and removes it. For relative
        paths (those not starting with '/'), it performs a depth-first search (DFS)
        to find the target node.

        Args:
            keychain_or_keychain_str (typing.Union[str, list[str]]): The path to the
                node to be removed. Can be a '/'-separated string or a list of
                strings.
            delete_key (bool, optional): Determines the removal behavior.
                - If `True` (default), the target key and all its children are
                  completely removed from the tree.
                - If `False`, the children are removed, but the key itself is
                  preserved with an empty string value.
            keychain_context (typing.Union[str, list[str], None], optional): An
                optional starting path that provides context for a relative
                `keychain` search. Defaults to None.

        Returns:
            typing.Union['Tree', typing.Any, None]: The removed value. This can be a
            scalar or a `Tree` object representing the removed branch. If the
            specified key is not found, the method does nothing and returns `None`.
        """

        _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str,list) \
                                                    else keychain_or_keychain_str.split('/')

        if _DEBUG.POP:
            ic()
            ic(self)
            ic(self.odict)
            ic(_keychain)
            ic(delete_key)
            ic(keychain_context)

        _trim_root = False
        if _keychain[-1] == '':
            _trim_root = True
            _keychain.pop()

        try:
            _absolute_keychain,_= self.dfs(_keychain,keychain_context)
            if _DEBUG.POP:
                ic(_absolute_keychain)
        except KeyError:
            #???? silently fail???
            return

        _value = self.get(keychain_or_keychain_str)
        if _DEBUG.POP:
            ic(_value)

        _new_tree = Tree()
        if not delete_key:
            _new_tree.get(keychain_or_keychain_str,'')


        for __keychain,__value in self.flatten():
            if _DEBUG.POP:
                ic(__keychain)

            if __keychain == _absolute_keychain:
                # ic(__keychain)
                # ic(_absolute_keychain)
                if not delete_key:
                    if _DEBUG.POP:
                        _msg = f'leaving keychain {keychain_or_keychain_str}'
                        ic(_msg)
                    _new_tree.get(keychain_or_keychain_str,'')
                else:
                    if _DEBUG.POP:
                        _msg = f'deleting keychain {keychain_or_keychain_str}'
                        ic(_msg)
                continue

            elif __keychain[:len(_absolute_keychain)] == _absolute_keychain:
                if not delete_key:
                    if _DEBUG.POP:
                        _msg = f'leaving keychain {keychain_or_keychain_str}'
                        ic(_msg)
                    _new_tree.get(keychain_or_keychain_str,'')
                continue
            else:
                _new_tree.get(__keychain,__value)

        # remove odict contents without breaking refs
        self.reset()
        self.overlay(_new_tree)

        if _DEBUG.POP:
            ic(self)
            ic(self.odict)
            if isinstance(_value,Tree):
                ic(_value.odict)
            else:
                ic(_value)
        return _value

    def flatten(self, relative: bool = False) -> list[tuple[list[str], typing.Any]]:
        """Creates a flat representation of the tree as a list of keychain-value pairs.

        Args:
            relative (bool, optional): Determines the path format for the keychains.
                - If `False` (default), keychains are absolute paths starting with an
                  empty string to represent the root (e.g., ['', 'a', 'b']).
                - If `True`, keychains are relative paths and do not start with an
                  empty string (e.g., ['a', 'b']). Defaults to False.

        Returns:
            list[tuple[list[str], typing.Any]]: A list of tuples, where each tuple
            contains a keychain (as a list of strings) and its corresponding value.

        Examples:
            >>> tree = Tree({'a': {'b': 'B_VAL'}, 'c': 'C_VAL'})
            >>> tree.flatten()
            [(['', 'a', 'b'], 'B_VAL'), (['', 'c'], 'C_VAL')]
            >>> tree.flatten(relative=True)
            [(['a', 'b'], 'B_VAL'), (['c'], 'C_VAL')]
        """
        if _DEBUG.FLATTEN:
            ic()
        _data = []
        def _val(node,keychain):
            if not relative:
                _data.append(([''] + keychain,node))
            else:
                _data.append((keychain,node))

        self.visit(value_process=_val)
        return _data

    def oget(self,keychain_or_keychain_str,value):
        """So we can use overlay like get in TreeState"""
        if isinstance(value,Tree):
            return self.overlay(value,keychain_or_keychain_str)
        return self.get(keychain_or_keychain_str,value)

    def overlay(self, tree: 'Tree', keychain_or_keychain_str: typing.Union[str, list[str]] = []) -> 'Tree':
        """Merges the contents of another Tree into the current one.

        This method overlays the key-value pairs from the source `tree` onto the
        current tree in-place. If a key already exists in the current tree, its
        value will be overwritten with the value from the source tree.

        Args:
            tree (Tree): The source `Tree` object to overlay.
            keychain_or_keychain_str (typing.Union[str, list[str]], optional): A base
                keychain to apply the overlay. If provided, the source tree will be
                merged at this location. If omitted, the merge happens at the root.
                Defaults to [].

        Returns:
            Tree: The modified `Tree` instance (`self`), allowing for method chaining.

        Examples:
            >>> base_tree = Tree({'a': 1, 'b': {'c': 2}})
            >>> overlay_tree = Tree({'b': {'d': 3}, 'e': 4})
            >>> base_tree.overlay(overlay_tree)
            >>> print(base_tree.dump())
            a: 1
            b:
              d: 3
            e: 4

            >>> target_tree = Tree({'app': {'config': {'version': '1.0'}}})
            >>> new_settings = Tree({'debug': True, 'version': '1.1'})
            >>> target_tree.overlay(new_settings, 'app/config')
            >>> print(target_tree.dump())
            app:
              config:
                debug: true
                version: '1.1'
        """
        _keychain = copy(keychain_or_keychain_str) if isinstance(keychain_or_keychain_str,list) \
                                                    else keychain_or_keychain_str.split('/')

        _relative = False
        if keychain_or_keychain_str:
            _relative = True

        for __keychain,__value in tree.flatten(_relative):
            if _DEBUG.OVERLAY:
                _msg = f"setting {'/'.join(_keychain + __keychain)} to {__value}"
                ic(_msg)
                ic(type(__value))

            self.get(_keychain + __keychain,__value)
        return self

    def print(self):
        """Construct a printable string representation of the tree.

        Returns:
            String: A printable indented string representing the tree
        """

        TAB='  '
        _n = 0
        _output = []
        def _pre(node,keychain):
            if not keychain: return
            nonlocal _n,TAB,_output
            _output.append(_n*TAB+keychain[-1]+':')
            _n += 1

        def _post(node,keychain):
            if not keychain: return
            nonlocal _n,TAB
            _n -= 1

        def _value(value, keychain):
            nonlocal _n,TAB
            _output.append(_n*TAB+keychain[-1]+':')

            if value is None: return
            if keychain[-1][0] == '_':
                # a hidden value
                _output.append((_n+1)*TAB+'...')
                return
            if isinstance(value, list):
                # no multiline values in lists
                for _value in value:
                    _output.append((_n+1)*TAB+str(_value))
            else:
                # possible multiline string print
                _output.append((_n+1)*TAB+str(value).replace('\n', '\n' + (_n + 1) * TAB))

        self.visit(_pre, _post, _value)
        return '\n'.join(_output)

    def stringify(self):
        """Ensure the leaf values in the tree are strings.

        Returns:
            String: A Tree object with all leaf values guaranteed to be stings.
        """
        if _DEBUG.STRINGIFY:
            ic()
        _data_tree = Tree()
        def _stringify(value,keychain):
            if _DEBUG.STRINGIFY:
                ic(keychain)
                ic(value)
            if isinstance(value,list):
                _data_tree.get([''] + keychain, list(map(str, value.decode() if isinstance(value,bytes) else value)))
            else:
                _data_tree.get([''] + keychain, str(value.decode() if isinstance(value,bytes) else value))

        self.visit(value_process=_stringify)
        return _data_tree

    def dump(self,stream=None,Dumper=yaml.Dumper,**kwds):
        """Serializes the Tree object into a YAML formatted string.

        This method preserves the order of keys from the original tree in the
        YAML output. It also ensures that all values within the tree are
        converted to their string representation before being serialized.

        Args:
            stream (typing.IO, optional): A file-like object (e.g., a file
                handle opened in write mode). If provided, the YAML output
                will be written to this stream. Defaults to None.
            Dumper (yaml.Dumper): The yaml.Dumper class to be used for the
                serialization process.
            **kwds: Additional keyword arguments that will be passed directly
                to the underlying yaml.dump function.

        Returns:
            str or None: If `stream` is None (the default), the method
            returns the YAML output as a string. If `stream` is provided,
            the method writes to the stream and returns None.
        """
        _data_tree = self.stringify()

        class OrderedDumper(Dumper):
            pass

        def _dict_representer(dumper, data):
            return dumper.represent_mapping(
                yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                data.items())

        OrderedDumper.add_representer(OrderedDict, _dict_representer)
        return yaml.dump(_data_tree.odict, stream, OrderedDumper, **kwds)

    @classmethod
    def _load(cls, stream):
        class OrderedLoader(yaml.BaseLoader):
            pass
        def construct_mapping(loader, node):
            return OrderedDict(loader.construct_pairs(node))

        OrderedLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            construct_mapping)
        return yaml.load(stream,OrderedLoader)

    @classmethod
    def load(cls,stream):
        """Constructs a new Tree instance from a YAML source.

        This factory @classmethod parses a YAML source, preserving the order of
        keys from the input.

        Args:
            stream (typing.Union[typing.IO, str]): The source of the YAML data.
                This can be a file-like object (e.g., from an open file) or a
                string containing YAML markup.

        Returns:
            Tree: A new Tree instance populated with the data from the stream.
        """
        return cls(Tree._load(stream))



from ..constants import *
from ..tree import *
from ..YAMLator import YAMLator
from .States import *
from .Observables import *


from . import DEBUG

class StateEvaluator(YAMLator):
    """ _post_evaluate,_pre_evaluate and _value_evaluate all have access to the state
        which is updated upon entering and exiting a node of nodes, or a node containing a value.
    """

    def __init__(self, tree, state,**kwargs):
        # if a subclass of this calls copy() it needs to call the constructor with root_dir=self.root_dir
        super().__init__(tree,**kwargs)
        self.state = state

    def read_state(self,observable):
        try:
            return list(filter(lambda x: isinstance(x, observable),
                               self.state.pre_observables + self.state.value_observables)).pop().read(self.state)
        except IndexError:
            return None

    def aggregate_state(self,observable):
        # we call laggretate() and daggregate() on self, an evaluator, not self.state like in read_state
        try:
            return list(filter(lambda x: isinstance(x, observable),
                               self.state.pre_observables + self.state.value_observables)).pop().read(self)
        except IndexError:
            return None

    def _pre_evaluate(self):
        pass

    def _post_evaluate(self):
        pass

    def _value_evaluate(self):
        pass

    def reset_state(self,*keychains_to_preserve):
        self.state.reset(*keychains_to_preserve)

    def evaluate(self,reverse=False):
        def _pre(node, keychain):
            if DEBUG.StateEvaluator:
                ic()
            self.state.pre_update(node,keychain)
            self._pre_evaluate()

        def _val(value,keychain):
            if DEBUG.StateEvaluator:
                ic()
            self.state.value_update(value,keychain)
            self._value_evaluate()

        def _post(node, keychain):
            self._post_evaluate()
            self.state.post_update(node, keychain)

        self.visit(_pre, _post, _val, reverse=reverse)
        return self



    def laggregate(self,property_name):
        '''aggregate all properties into a single list'''
        class _PropertyAggregator(RegexObservable):
            def __init__(self, property):
                self.keychain_regex = rf'^{REGEXES.KEYCHAIN}({property})$'
                super().__init__()
                self.property = property

            def _create_observation(self, node, *parameters):
                assert parameters[0] == self.property
                return self.property, node

        return StateEvaluator(
            self.state,
            AggregateState(
                value_observables=(_PropertyAggregator(property_name),),
                initial_state=Tree(
                    OrderedDict({
                        property_name: []
                    })
                ),
            )
        ).evaluate().state.get(property_name)

    def daggregate(self,property_name):
        '''aggregate all property dictionaries into a single dict'''
        class _PropertyDaggregator(RegexObservable):
            def __init__(self, property):
                self.keychain_regex = rf'^{REGEXES.KEYCHAIN}({property})$'
                super().__init__()
                self.property= property

            def _create_observation(self, node, *parameters):
                # if DEBUG.StateEvaluator:
                #     ic()
                #     ic(parameters)
                #     ic(node)
                assert parameters[0] == self.property
                if isinstance(node,list):
                    return self.property, reduce(lambda x, y: {**y, **x}, filter(lambda x: x is not None, node))
                else:
                    return self.property,node

        return StateEvaluator(
            self.state,
            DAggregateState(
                value_observables=(_PropertyDaggregator(property_name),),
                initial_state=Tree(
                    OrderedDict({
                        property_name: {}
                    })
                )
            )
        ).evaluate().state.get(property_name)



class ObjectFunctor:
    def __init__(self, function, *args, **kargs):
        assert callable(function), "function should be a callable obj"
        self._function = function
        self._args = args
        self._kargs = kargs

    def __call__(self, *args, **kargs):
        """call function"""
        _args = list(self._args)
        _args.extend(args)
        _kargs = self._kargs.copy()
        _kargs.update(kargs)
        return self._function(*_args, **_kargs)


class ObjectDB:
    def _register(self, name, constructor, *args, **kargs):
        """register a constructor"""
        _args = [constructor]
        _args.extend(args)
        setattr(self, name, ObjectFunctor(*_args, **kargs))




from ..constants import *
from .ValueTransformer import *

from . import DEBUG

class AtTransformerException(Exception):
    pass


class AtTransformer(ValueTransformer):
    name = 'transform_ats'
    match_regex = \
        f"{re.escape(KEY_OR_KEYCHAIN_OP)}{re.escape(KEYCHAIN_LEFT_BOUND)}?" + \
        f"{REGEXES.AT}{re.escape(KEYCHAIN_RIGHT_BOUND)}?"
    extract_regex = \
        f"{re.escape(KEY_OR_KEYCHAIN_OP)}({re.escape(KEYCHAIN_LEFT_BOUND)}?)" + \
        f"({REGEXES.AT_EXP})({re.escape(KEYCHAIN_RIGHT_BOUND)}?)"

    def __init__(self, odict_or_tree=None, **kwargs):
        odict_or_tree = OrderedDict() if odict_or_tree is None else odict_or_tree
        super(AtTransformer, self).__init__(odict_or_tree)
        if DEBUG.AtTransformer:
            ic()

    def _match(self, line):
        _m = re.match(f"(.*?)({self.match_regex})(.*)",str(line))
        if _m:
            if DEBUG.AtTransformer:
                ic()
                ic(line)
                ic(_m.groups())
        return _m

    def _extract(self, token):
        _m = re.match(f"{self.extract_regex}",str(token))
        if _m:
            if DEBUG.AtTransformer:
                ic()
                ic(token)
                ic(_m.groups())
        return _m

    def _transform(self, parameters, keychain):
        if  DEBUG.AtTransformer:
            ic(parameters)
            ic(keychain)

        _l_brace, _special_keychain_param, _sign, _index, _r_brace = parameters

        _transformed_token = None

        if DEBUG.AtTransformer:
            ic(_l_brace)
            ic(_special_keychain_param)
            ic(_sign)
            ic(_index)
            ic(_r_brace)

        # TODO: Note that if we are making a keychain based on @
        # and there is a variable in the keychain, we need to 'reverse escape' the
        # path: key1/))key2//key3 , not key1/))key2/key3 ; we don't want ))key2/key3 to parse as a keychain !
        if any(map(is_variable_token,keychain)):
            _old_keychain = copy(keychain)
            keychain = []
            for _key in _old_keychain:
                if is_variable_token(_key):
                    keychain.append(_key + '/')
                else:
                    keychain.append(_key)

        if _index:
            _index = int(_index)
            if _sign:
                _transformed_token = '/'.join(keychain[:-_index])
            else:
                _transformed_token = '/'.join(keychain[:_index])
        else:
            _transformed_token = '/'.join(keychain)

        if _l_brace == '' and _r_brace == '':
            _transformed_token = _transformed_token.split('/')[-1]

        if DEBUG.AtTransformer:
            ic(_transformed_token)

        return _transformed_token

class AtTransformerUtility(AtTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



class DEBUG:
    Transformer = False
    KeyChainTransformer = False

    ValueTransformer = False
    KeyTransformer = False

    PathValueTransformer = False
    YAMLTransformer = False
    PythonTransformer = False
    PlainTextTransformer = False

    AtTransformer = False

    IfTransformer = False
    IfKeyTransformer = False

    BangTransformer = False
    ImportTransformer = False


from .Transformer import Transformer,TransformerException
from .KeyTransformer import KeyTransformer,KeyTransformerUtility
from .YAMLTransformer import YAMLTransformer,YAMLTransformerUtility
from .AtTransformer import AtTransformer,AtTransformerUtility
from .IfTransformer import IfTransformer,IfKeyTransformer,IfKeyTransformerUtility,IfTransformerUtility
from .BangTransformer import BangTransformerException,BangTransformer,BangTransformerUtility
from .ImportTransformer import ImportTransformer,ImportTransformerUtility
from .PlainTextTransformer import PlainTextTransformer,PlainTextTransformerUtility

# to eliminate circular imports due to the use of ValueTransformer in ValueValidator
# eventually all utilities will be here
from .ValueTransformer import ValueTransformer
from ..validators.ValueValidator import ValueValidator
class ValueTransformerUtility(ValueTransformer,ValueValidator):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)


from ..constants import *
from .ValueTransformer import *

from . import DEBUG

class PathValueTransformerException(Exception):
    pass


class PathValueTransformer(ValueTransformer):
    root_dir = None
    ext = None
    # the ? after # is critical. But why?
    match_regex = rf'(?:{REGEXES.POSIX_RELATIVE}|{REGEXES.POSIX_ABSOLUTE})#?{REGEXES.KEYCHAIN}?(?:{REGEXES.KEY})?/?'
    extract_regex = r'^([^#]*?)#(.*)$'

    def __init__(self, odict_or_tree):
        # if not hasattr(odict_or_tree, 'root_dir'):
        #     raise PathValueTransformerException
        # self.root_dir = odict_or_tree.root_dir
        super(PathValueTransformer,self).__init__(odict_or_tree)


    def select(self, stream, selector):
        '''intpret a selctor string to select bytes from stream'''
        raise NotImplemented

    def _do_not_evaluate(self,value,keychain):
        # don't try to pathtransform on multi=line strings
        if isinstance(value,str) and len(value.split('\n')) > 1:
            return True

    def _match(self, line):
        # TODO: Note this
        # we have to avoid matching the internal structure of a URI somehow...
        _m = re.match(rf'([^:]*?)({self.match_regex})(.*)',str(line))
        if _m:
            if DEBUG.PathValueTransformer:
                ic()
                ic(line)
                ic(_m.groups())
        return _m

    def _extract(self, token):
        if DEBUG.PathValueTransformer:
            ic()
            ic(token)
        _m = re.match(self.extract_regex,str(token))
        if _m:
            if DEBUG.PathValueTransformer:
                ic()
                ic(token)
                ic(_m.groups())
        # else:
            # if DEBUG.PathValueTransformer:
            #     _msg = f'No path selector match for {token}!'
            #     ic(_msg)
        return _m

    def _transform(self, parameters, keychain):
        _path_str,_selector = parameters
        _path = pathlib.Path(_path_str)
        _ext = _path.name.split('.')[-1]
        if _ext == _path.name:
            _ext = None

        if DEBUG.PathValueTransformer:
            ic(_path)
            ic(_selector)
            ic(self.ext)

        if _ext != self.ext:
            if DEBUG.PathValueTransformer:
                _msg = f'wrong ext {_ext}, not {self.ext}'
                ic(_msg)
            return

        # only transform paths with the right extension!!!
        # if self.ext and _path.name.split('.')[-1] != self.ext:
        #     if DEBUG.PathValueTransformer:
        #         _msg = f'wrong extension {_path.name.split(".")[-1]} not {self.ext}'
        #         ic(_msg)
        #     return None
        # elif self.ext is None and _path.name.split('.')[-1] != _path.name :
        #     if DEBUG.PathValueTransformer:
        #         _msg = f'Found a plain text file to read'
        #         ic(_msg)


        # we handle absolute paths in yaml selectors
        if pathlib.Path(_path_str).is_absolute():
            _value_path = _path
        else:
            # there's always a root_dir set in the superclass
            _value_path = self.root_dir.joinpath(_path)

        try:
            with _value_path.open('r') as f:
                _new_value = self.select(f, _selector)
        except FileNotFoundError as e:
            if DEBUG.PathValueTransformer:
                ic(e)
            _new_value = None
        except KeyError as e:
            if DEBUG.PathValueTransformer:
                ic(e)
            _new_value = None
        return _new_value



from ..constants import *

class ValidationIssue(Enum):
    """Enumeration for all possible validation issues."""
    CIRCULAR_DEPENDENCY = "Circular dependency found: {path}"
    UNDEFINED_VARIABLE = "Warning: Undefined variable '{variable}' is found."
    UNUSED_VARIABLE = "Warning: Unused variable '{variable}' is defined but never used."


@dataclass
class ValidationResult:
    """Represents a single validation issue found in a tree."""
    issue_type: ValidationIssue
    message: str

    def __str__(self) -> str:
        """Returns the human-readable message for this validation issue."""
        return self.message


from .constants import *
from .tree import *

class UnOrderedTreeException(Exception):
    pass

class UTree(Tree):
    def __init__(self, odict_or_dict_or_tree=None):
        super().__init__()
        if odict_or_dict_or_tree is None:
            self.odict = dict()
        elif isinstance(odict_or_dict_or_tree, dict):
            self.odict = odict_or_dict_or_tree
        elif issubclass(odict_or_dict_or_tree.__class__, UTree):
            self.odict = odict_or_dict_or_tree.odict
        else:
            raise UnOrderedTreeException('Cannot create tree')

    def visit(self,pre_process=lambda x, y:None, post_process=lambda x, y:None, value_process=lambda x, y:None,reverse=False,entry_keychain=None):
        """Reverse visting is not supported for UTree objects."""
        if reverse:
            raise UnOrderedTreeException('Reverse visiting is not supported !')
        if entry_keychain:
            raise UnOrderedTreeException('Tree entry keychains are not supported !')

        self._visit(self.odict, pre_process, post_process, value_process, None)

    visit.__doc__ += '\n' + Tree.visit.__doc__

    def _visit(self, node, pre_process=lambda x, y:None, post_process=lambda x, y:None, value_process=lambda x, y:None,keychain=None,entry_keychain=None,process=None):
        keychain = [] if keychain is None else keychain

        if isinstance(node,dict) or isinstance(node,OrderedDict):
            pre_process(node,copy(keychain))
            for _child_key in node.keys():
                keychain.append(_child_key)
                self._visit(node.get(_child_key), pre_process, post_process, value_process, copy(keychain))
                keychain.pop()
            post_process(node,copy(keychain))

        else:
            value_process(node,keychain)

    def copy(self):
        '''return a copy of the tree'''
        return UTree(deepcopy(self.odict))





from ..constants import *
from .AbstractState import  *

from . import DEBUG

class AggregateState(AbstractState):
    def pre_update(self,node,keychain):
        if DEBUG.AggregateState:
            ic()
            ic(keychain)
        for _observation in self.pre_observables:
            _result = _observation.observe(node,keychain)
            if _result:
                if _observation.aggregate:
                    self.aget(*_result)
                else:
                    # TODO: observables can skip aggregating with this flag
                    self.get(*_result)

            if DEBUG.AggregateState:
                ic(_observation.__class__)
                ic(_result)

    def value_update(self,value,keychain):
        if DEBUG.AggregateState:
            ic()
            ic(keychain)
        for _observation in self.value_observables:
            _result = _observation.observe(value, keychain)
            if _result:
                if _observation.aggregate:
                    self.aget(*_result)
                else:
                    self.get(*_result)

            if DEBUG.AggregateState:
                ic(_observation.__class__)
                ic(_result)

    # this never pops any state created by value_update
    def post_update(self,node,keychain):
        if DEBUG.AggregateState:
            ic()
            ic(keychain)
        for _observation in self.post_observables:
            _result = _observation.observe(node, keychain)
            if _result:
                _state_keychain, _value = _result
                if DEBUG.AggregateState:
                    ic(_state_keychain)
                    ic(_value)
                try:
                    _values = self.get(_state_keychain)
                    if DEBUG.AggregateState:
                        ic(_values)
                except KeyError:
                    # _state_keychain has already been popped ?
                    if DEBUG.AggregateState:
                        _msg = f'{_state_keychain} not found; ignoring'
                        ic(_msg)
                    # return
                    continue
                if isinstance(_values,list) and _values and _observation.aggregate:
                    if not isinstance(_value,list):
                        _value = [_value]
                    try:
                        for __value in _value:
                            _popped_value = _values.pop(_values.index(__value))
                            assert _popped_value in _value
                            _new_value = self.get(_state_keychain,_values)
                            if DEBUG.AggregateState:
                                ic(_new_value)
                    except ValueError as e:
                        if DEBUG.AggregateState:
                            ic()
                            ic(self.__class__)
                            ic(e)
                        continue
                else:
                    self.pop(_state_keychain)

            if DEBUG.AggregateState:
                ic(_observation.__class__)
                ic(_result)


class DAggregateState(AbstractState):
    def pre_update(self,node,keychain):
        for _observation in self.pre_observables:
            _result = _observation.observe(node,keychain)
            if _result:
                self.daget(*_result)
                if DEBUG.DAggregateState:
                    ic()
                    ic(keychain)
                    ic(self.print())

    def value_update(self,value,keychain):
        for _observation in self.value_observables:
            _result = _observation.observe(value, keychain)
            if _result:
                self.daget(*_result)
                if DEBUG.DAggregateState:
                    ic()
                    ic(keychain)
                    ic(self.print())

    def post_update(self,node,keychain):
        for _observation in self.post_observables:
            _result = _observation.observe(node, keychain)
            if _result:
                _state_keychain, _value = _result
                try:
                    _values = self.get(_state_keychain)
                except KeyError:
                    return
                if isinstance(_values,dict) and _values:
                    _new_values = [_values.pop(_k) for _k in _value.keys()]
                    self.get(_state_keychain,_new_values)
                else:
                    self.pop(_state_keychain)
                if DEBUG.AggregateState:
                    ic()
                    ic(keychain)
                    ic(_state_keychain)


class TreeState(AbstractState):
    def pre_update(self,node,keychain):
        for _observation in self.pre_observables:
            _result = _observation.observe(node,keychain)
            if _result:
                # we need to use an overlay type get here, I think
                self.oget(*_result)
                if DEBUG.TreeState:
                    ic()
                    ic(keychain)
                    ic(_result[0])
                    ic(self.keys())
                    # ic(self.print())

    def value_update(self,value,keychain):
        for _observation in self.value_observables:
            _result = _observation.observe(value, keychain)
            if _result:
                self.oget(*_result)
                if DEBUG.TreeState:
                    ic()
                    ic(keychain)
                    ic(_result[0])
                    ic(self.keys())
                    # ic(self.print())
    def post_update(self,node,keychain):
        if DEBUG.TreeState:
            ic()
            ic(keychain)
        for _observation in self.post_observables:
            _result = _observation.observe(node, keychain)
            if _result:
                _state_keychain, _value = _result
                if DEBUG.TreeState:
                    ic(_state_keychain)
                    ic(_value)
                try:
                    _values = self.get(_state_keychain)
                    if DEBUG.TreeState:
                        ic(_values)
                except KeyError:
                    # _state_keychain has already been popped ?
                    if DEBUG.TreeState:
                        _msg = f'{_state_keychain} not found; ignoring'
                        ic(_msg)
                    # return
                    continue
                self.pop(_state_keychain)

            if DEBUG.TreeState:
                ic(_observation.__class__)
                ic(_result)



from ..constants import *
from ..tree import *
from .KeyTransformer import *
from .YAMLTransformer import *
from .ValueTransformer import *

from . import DEBUG

class ImportTransformer(KeyTransformer):
    name = 'transform_imports'
    match_regex = rf"{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.IMPORT}"
    extract_regex = rf"({re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.IMPORT_EXP})"

    def __init__(self, odict_or_yamlator_or_tree=None, context_tree=None, allow_tree_subs=False, root_dir=None):
        if hasattr(odict_or_yamlator_or_tree, 'root_dir') and root_dir is None:
            self.root_dir = odict_or_yamlator_or_tree.root_dir
        else:
            root_dir = pathlib.Path('.') if root_dir is None else root_dir
            self.root_dir = pathlib.Path(root_dir).absolute()

        super(ImportTransformer,self).__init__(odict_or_yamlator_or_tree, context_tree)

    def _replace_node_key(self, node, node_key, _transformed_key):

        if DEBUG.KeyTransformer:
            ic(node_key)
            ic(_transformed_key)
        _node_copy = deepcopy(node)
        _key_index = list(_node_copy.keys()).index(node_key)

        for _key in list(_node_copy.keys())[:_key_index]:
            if DEBUG.KeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

        if DEBUG.KeyTransformer:
            _msg = f'replacing {node_key} with {_transformed_key}'
            ic(_msg)

        _value = node.pop(node_key)
        if DEBUG.KeyTransformer:
            ic(_value)

        for _k,_v in _transformed_key.items():
            node[_k] = _v
        # node[_transformed_key] = _value

        for _key in list(_node_copy.keys())[_key_index + 1:]:
            if DEBUG.KeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

    def _transform(self, parameters, keychain):
        _node_key, = parameters
        if _node_key[-1] == '/':
            _node_key = _node_key[:-1]

        _selector_value = self.get(keychain + list(parameters) + [''])
        if DEBUG.ImportTransformer:
            ic(keychain)
            ic(parameters)
            ic(_selector_value if not isinstance(_selector_value,Tree) else _selector_value.print())

        if not isinstance(_selector_value,Tree):
            _input_tree = Tree(OrderedDict({_node_key:_selector_value}))
            ValueTransformer(_input_tree,context_tree=self,allow_tree_subs=False).evaluate()

            YAMLTransformer(_input_tree,root_dir=self.root_dir).evaluate()

            _transformed_value = _input_tree.get(_node_key + '/')
            # TODO: this is broken; why cannot we pass a keychain string??
            # _transformed_value = _input_tree.get([_node_key,''])
            _path,_selector = _selector_value.split('#')
            _new_root_dir = self.root_dir.joinpath(pathlib.Path(_path).parent)
            ImportTransformer(_transformed_value,root_dir=_new_root_dir).evaluate()
        else:
            _transformed_value = _selector_value
            ImportTransformer(_transformed_value,root_dir=self.root_dir).evaluate()

        if isinstance(_transformed_value,Tree):
            if DEBUG.ImportTransformer:
                ic(_transformed_value.print())

            _transformed_value = _transformed_value.odict

        return _transformed_value

    def _pre_evaluate(self, node, keychain):
        if DEBUG.KeyTransformer:
            ic()
            ic(keychain)
            ic(node.keys())

        for _node_key in copy(list(node.keys())):
            _tmp_key = _node_key

            _tokenized_key = self._tokenize(_tmp_key)
            if DEBUG.KeyTransformer:
                ic(_tokenized_key)

            _transformed_token = None
            for _token in _tokenized_key:
                _transformed_token = _token
                _match = self._extract(_token)
                if _match is not None:
                    _transformed_token = self._transform(_match.groups(), keychain)

                if _transformed_token is None:
                    _transformed_token = _token
                elif isinstance(_transformed_token, list):
                    raise Exception(f'Are you really trying to turn a list into a key?')
                if DEBUG.KeyTransformer:
                    ic(_transformed_token)


            if isinstance(_transformed_token,OrderedDict):
                self._replace_node_key(node,_node_key,_transformed_token)


class ImportTransformerUtility(ImportTransformer):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)


from ..constants import *
from ..tree import *
from .Transformer import *

from . import DEBUG

class KeyChainTransformer(Transformer):
    match_regex = \
        f"{re.escape(KEY_OR_KEYCHAIN_OP)}{re.escape(KEYCHAIN_LEFT_BOUND)}?" + \
        f"{REGEXES.KEYCHAIN}{REGEXES.KEY}/?{re.escape(KEYCHAIN_RIGHT_BOUND)}?" + \
        f"(?:{re.escape('[')}[^]]*{re.escape(']')})?"

    extract_regex = \
        f"{re.escape(KEY_OR_KEYCHAIN_OP)}{re.escape(KEYCHAIN_LEFT_BOUND)}?" + \
        f"({REGEXES.KEYCHAIN}{REGEXES.KEY}/?){re.escape(KEYCHAIN_RIGHT_BOUND)}?" + \
        f"(?:{re.escape('[')}({REGEXES.DATA_INDEX}){re.escape(']')})?"

    def __init__(self, odict_or_tree=None, context_tree=None, allow_tree_subs=True):
        # super(KeyChainTransformer,self).__init__(odict_or_tree)
        super().__init__(odict_or_tree)
        self.context = context_tree
        self.allow_tree_subs = allow_tree_subs

    def _match(self, line):
        _m = re.match(rf"^(.*?)({self.match_regex})(.*)$", str(line))

        if _m:
            if DEBUG.KeyChainTransformer:
                ic(str(line))
                ic(_m.groups())
        return _m

    def _extract(self, token):
        _m = re.match(self.extract_regex,str(token))
        if _m:
            if DEBUG.KeyChainTransformer:
                ic(token)
                ic(_m.groups())

        return _m

    def _transform(self, parameters, keychain):
        if DEBUG.KeyChainTransformer:
            ic(parameters)

        _keychain_param, _data_index = parameters

        _local_context = self if not self.context else self.context

        _slice = None
        _local_context_keychain = None

        if _data_index:
            # slices take precedence so try to match first!
            _m = re.match(REGEXES.SLICE, _data_index)
            if _m:
                # we found a slice
                _slice = _data_index
                if DEBUG.KeyChainTransformer:
                    ic(_slice)

            else:
                # we found a local context
                _local_context_keychain = _data_index
                try:
                    if self.context:
                        _local_context = self.context.get(_local_context_keychain)
                    else:
                        _local_context = self.get(_local_context_keychain)

                    assert isinstance(_local_context,Tree)

                    if DEBUG.KeyChainTransformer:
                        ic(_local_context.print())
                except KeyError:
                    if DEBUG.KeyChainTransformer:
                        _msg = f'cannot find local context key {_local_context_keychain} in global context'
                        ic(_msg)
                    return None

            _substitution_value = None

        try:
            _substitution_value = _local_context.get(_keychain_param)
            # _substitution_value = _local_context.get(_local_context_keychain)
        except KeyError:
            if DEBUG.KeyChainTransformer:
                _msg = f'cannot find key {_keychain_param}'
                ic(_msg)
            return None

        if DEBUG.KeyChainTransformer:
            ic(_substitution_value)

        if isinstance(_substitution_value, OrderedDict):
            _substitution_value = str(Tree(_substitution_value))
            if DEBUG.KeyChainTransformer:
                _msg = 'found an ordereddict substitution'
                ic(_msg)

        # elif isinstance(_substitution_value, list):

        elif isinstance(_substitution_value, Tree):
            if DEBUG.KeyChainTransformer:
                _msg = 'found a tree substitution'
                ic(_msg)
            if not self.allow_tree_subs:
                _substitution_value = None

        if _slice is not None:
            if DEBUG.KeyChainTransformer:
                _msg = f'found a slice '
                ic(_msg)
            _substitution_value = eval(f"_substitution_value[{_slice}]")

        if DEBUG.KeyChainTransformer:
            ic(_substitution_value)

        return _substitution_value



from ..constants import *
from .PathValueTransformer import *

from . import DEBUG

class PlainTextTransformerException(Exception):
    pass


class PlainTextTransformer(PathValueTransformer):
    name = 'transform_plaintext'
    ext = None

    def __init__(self, odict_or_yamlator_or_tree, context_tree=None, allow_tree_subs=False,root_dir=None):

        if hasattr(odict_or_yamlator_or_tree, 'root_dir') and root_dir is None:
            self.root_dir = odict_or_yamlator_or_tree.root_dir
        else:
            root_dir = pathlib.Path('.') if root_dir is None else root_dir
            self.root_dir = pathlib.Path(root_dir).absolute()

        super(PlainTextTransformer,self).__init__(odict_or_yamlator_or_tree)

        if DEBUG.PlainTextTransformer:
            ic(self.__class__)
            ic(self.ext)
            ic(self.root_dir)
            # ic(self.print())


    def select(self, stream, keychain):
        if DEBUG.PlainTextTransformer:
            ic()
            ic(keychain)

        # let's avoid a syscall that's going to fail
        if any(map(is_variable_token,keychain.split('/'))):
            if DEBUG.PlainTextTransformer:
                _msg = f'found unsubbed vars'
                ic(_msg)
                ic(keychain)
            return None

        _stream_path = pathlib.Path(stream.name)

        _plain_text = stream.read()

        if DEBUG.PlainTextTransformer:
            ic(_stream_path)
            ic(_plain_text)

        return _plain_text

    # # for testing
    # def _match(self, line):
    #     ic(line)
    #     # TODO: Note this
    #     # we have to avoid matching the internal structure of a URI somehow...
    #     _m = re.match(rf'([^:]*?)({self.match_regex})(.*)',str(line))
    #     if _m:
    #         ic()
    #         ic(line)
    #         ic(_m.groups())
    #     return _m
    #
    # def _extract(self, token):
    #     ic()
    #     ic(token)
    #     _m = re.match(r'^([^#]*?)#(.*)$',str(token))
    #     if _m:
    #         ic()
    #         ic(token)
    #         ic(_m.groups())
    #     # else:
    #         # if DEBUG.PathValueTransformer:
    #         #     _msg = f'No path selector match for {token}!'
    #         #     ic(_msg)
    #     return _m

    # def _transform(self, parameters, keychain):
    #     ic()
    #     _path_str,_selector = parameters
    #     _path = pathlib.Path(_path_str)
    #     ic(_path)
    #     ic(_selector)
    #
    #     # only transform paths with the right extension!!!
    #     if self.ext and _path.name.split('.')[-1] != self.ext:
    #         _msg = f'wrong extension {_path.name.split(".")[-1]} not {self.ext}'
    #         ic(_msg)
    #         return
    #     else:
    #         _msg = f'Found a plain text file to read'
    #         ic(_msg)
    #
    #
    #     # we handle absolute paths in yaml selectors
    #     if pathlib.Path(_path_str).is_absolute():
    #         _value_path = _path
    #     else:
    #         if DEBUG.PlainTextTransformer:
    #             ic(self.root_dir)
    #         _value_path = self.root_dir.joinpath(_path)
    #
    #     try:
    #         with _value_path.open('r') as f:
    #             _new_value = self.select(f, _selector)
    #     except FileNotFoundError as e:
    #         ic(e)
    #         _new_value = None
    #     except KeyError as e:
    #         ic(e)
    #         _new_value = None
    #     if DEBUG.PlainTextTransformer:
    #         ic(_new_value)
    #
    #     return _new_value
    #
    #


class PlainTextTransformerUtility(PlainTextTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


from .tree import *
from .transformers import *
from .objdb import *
from .constants import *


DEFAULT_UTILITIES= {
    ImportTransformerUtility.name:ImportTransformerUtility,
    ValueTransformerUtility.name:ValueTransformerUtility,
    AtTransformerUtility.name:AtTransformerUtility,
    BangTransformerUtility.name:BangTransformerUtility,
    YAMLTransformerUtility.name:YAMLTransformerUtility,
    KeyTransformerUtility.name:KeyTransformerUtility,
    IfKeyTransformerUtility.name:IfKeyTransformerUtility,
    IfTransformerUtility.name:IfTransformerUtility,
    PlainTextTransformerUtility.name:PlainTextTransformerUtility,
}


class _DEBUG:
    YAMLator = False
    CONFIG_ATTRS = False
    VALIDATE = False
    TRANSFORM = False


class YAMLatorException(Exception):
    pass


class YAMLatorObjectDB(ObjectDB):
    regexes = None

    def __init__(self):
        super().__init__()
        self.regexes = {}


    def add_object(self, key_type, key_type_regex, constructor, *args, **kwargs):
        self._register(key_type, constructor, *args, **kwargs)
        self.regexes[key_type] = key_type_regex

    def get_object(self,key,value):
        try:
            return getattr(self,self.key_type(key))(value)
        except TypeError:
            return value

    def get_as_object(self,key):
        _value = self.get(key)
        try:
            return getattr(self,self.key_type(key))(_value)
        except TypeError:
            _msg = f"could not make a {self.key_types(key)} object from {_value}"
            raise YAMLatorException(f"could not make an ")

    def key_types(self):
        return list(self.regexes.keys())

    def key_type(self,key):
        _key_type = None
        for __key_type, _regex in self.regexes.items():
            _m = re.match(_regex, key)
            if _m:
                if _key_type is not None and _key_type != __key_type:
                    raise YAMLatorException('we matched two keys types for one key!')
                _key_type = __key_type
        return _key_type

    def register(self,object_class,attribute_name,*args,**kwargs):
        self._register(attribute_name,object_class,self,*args,**kwargs)

class YAMLator(YAMLatorObjectDB, Tree):
    """A powerful, high-level configuration processing engine.

    YAMLator provides a robust framework for managing and processing complex
    YAML-based configurations. It inherits from the `Tree` class, giving it a
    hierarchical, dictionary-like data structure that is easy to navigate.

    Its two primary features are:
    1.  A transformation engine that iteratively resolves macros, variables,
        imports, and conditionals. This process continues until the
        configuration is fully expanded and reaches a stable state (see the
        `transform` method).
    2.  A dynamic type system that automatically casts string values into rich
        Python objects (e.g., `pathlib.Path`, `bool`, `yarl.URL`) based on
        regular expression patterns matched against their corresponding keys.

    The primary use case is to load a templated YAML file and transform it
    into a fully resolved, type-safe, and easily accessible configuration
    object.

    Attributes:
        root_dir (pathlib.Path): The absolute path that serves as the base
            for resolving relative file paths during transformations, such as
            for `!import` directives.
        (UPPERCASE_ATTRIBUTES): After processing (e.g., by calling
            `set_config_attrs`), the instance is populated with additional
            uppercase attributes. These attributes correspond to keys in the
            YAML data and provide direct, type-casted access to the final
            configuration values.
    """

    @classmethod
    def load(cls, stream):
        """A class method factory for creating a new YAMLator instance from a stream.

        This method overrides the parent `Tree.load` method. Its primary purpose
        is to automatically determine the `root_dir` for the new instance, which
        is crucial for resolving relative file paths in directives like `!import`.

        The `root_dir` is determined based on the stream type:
        - If the stream is a file, `root_dir` is set to the file's directory.
        - If the stream is an `io.StringIO` instance, it uses a default path
          based on the current working directory.

        Args:
            stream (file-like object or io.StringIO): The source of the YAML
                data. This can be a file-like object (from an open file) or
                an `io.StringIO` instance for in-memory YAML content.

        Returns:
            YAMLator: A new `YAMLator` instance, initialized with the parsed
                data and the determined `root_dir`.
        """
        if isinstance(stream,io.StringIO):
            _root_dir = str(pathlib.Path('.').parent)
        else:
            _root_dir = str(pathlib.Path(stream.name).parent)
        # return YAMLator(cls._load(stream),_root_dir)
        return cls(cls._load(stream),_root_dir)


    def __init__(self, odict_or_yamlator_or_tree=None, root_dir=None):
        """Initializes a new YAMLator instance.

        Args:
            odict_or_yamlator_or_tree (OrderedDict | YAMLator | Tree | str | None, optional):
                Provides the initial data for the object. It can be one of
                several types:
                - OrderedDict: To initialize from a pre-existing dictionary.
                - YAMLator or Tree: To initialize from another instance,
                  effectively creating a copy.
                - str: To initialize by parsing a string of YAML content.
                - None: To create an empty YAMLator instance.
                Defaults to None.
            root_dir (pathlib.Path or str, optional): The base directory used
                to resolve relative file paths found within the YAML content
                (e.g., for `!import` directives). If not provided, it will be
                inferred from the input data source (if it has a `root_dir`
                attribute) or will default to the current working directory.
                Defaults to None.
        """
        # this order is critical
        YAMLatorObjectDB.__init__(self)
        Tree.__init__(self, odict_or_yamlator_or_tree)
        if hasattr(odict_or_yamlator_or_tree, 'root_dir') and root_dir is None:
            self.root_dir = odict_or_yamlator_or_tree.root_dir
        else:
            root_dir = pathlib.Path('.') if root_dir is None else root_dir
            self.root_dir = pathlib.Path(root_dir).absolute()

        # ensure custom objdb has the default types
        for _key_type, _data in DEFAULT_KEY_TYPES.items():
            _key_regex, _object_constructor_data = _data
            self.add_object(_key_type, _key_regex, *_object_constructor_data)

        # register the default utilities
        for _attribute_name,_transformer_class in DEFAULT_UTILITIES.items():
            self.register(_transformer_class,_attribute_name)


    def transform(self,methods=None,context_tree=None,allow_tree_subs=True):
        """The main processing engine for the YAMLator object.

        This method repeatedly applies a suite of transformation utilities to the
        data. The process continues in a loop until the data reaches a "stable
        state" or "fixed point," meaning a full pass of all transformers causes
        no further changes.

        This iterative approach is necessary to correctly resolve interdependent
        operations, such as importing a file and then substituting variables
        that were defined within it.

        Args:
            methods (list[str], optional): A sequence of strings, where each
                string is the name of a transformer method to run. If None,
                all default transformers are used. Defaults to None.
            context_tree (Tree | YAMLator, optional): An optional Tree or
                YAMLator object that provides an external context for lookups
                (e.g., for resolving variables not defined in the current
                tree). Defaults to None.
            allow_tree_subs (bool, optional): A boolean flag passed to the
                transformers to enable or disable a specific substitution
                feature. Defaults to False.

        Returns:
            None: The method modifies the YAMLator instance in-place.
        """
        # apply all default utilities until no more change in self.odict
        methods = methods if methods is not None else tuple(DEFAULT_UTILITIES.keys())
        _methods = []
        if _DEBUG.TRANSFORM:
            ic(methods)

        for _method in methods:
            _methods.append(getattr(self, _method))
            if _DEBUG.TRANSFORM:
                ic(_methods[-1]().name)

        _old_odict = self.copy().odict
        if _DEBUG.TRANSFORM:
            ic(self.odict)
        for _method_utility in _methods:
            _method_utility(context_tree=context_tree, allow_tree_subs=allow_tree_subs).evaluate()

        while self.odict != _old_odict:
            if _DEBUG.TRANSFORM:
                ic(self.odict)
            _old_odict = self.copy().odict
            for _method_utility in _methods:
                _method_utility(context_tree=context_tree, allow_tree_subs=allow_tree_subs).evaluate()

    def get(self,keychain_or_keychain_str,value=None):
        """
        Wraps Tree.get() to return a YAMLator instance.

        This method follows the exact same semantics as the parent method,
        but ensures the return type is a YAMLator object for chainable
        operations.
        """
        # AND transfer registered utilities and objdb
        _tmp_value = super(self.__class__,self).get(keychain_or_keychain_str,value)
        if isinstance(_tmp_value,Tree):
            _yt = self.__class__(
                _tmp_value,
                root_dir=self.root_dir)
            _yt.set_config_attrs()
            return _yt
        else:
            return _tmp_value
    # Append the parent's detailed docstring after the class is defined
    get.__doc__ += '\n' + Tree.get.__doc__

    def copy(self):
        """Copies the YAMLator object.

        Returns:
             YAMLator: A deepcopy of the YAMLator object.
        """
        _yt = self.__class__(deepcopy(self.odict),root_dir=self.root_dir)
        _yt.set_config_attrs()
        return _yt

    def merge(self, yaml_path_selector, keychain_str=None, relative=False):
        """Merges data from an external YAML file into the current instance.

        This method imports data from a specified YAML file and merges it into
        the current YAMLator tree in-place. It allows for selecting a specific
        portion of the external file to merge.

        Args:
            yaml_path_selector (str): A string specifying the source file and an
                optional data selector, formatted as "path/to/file.yaml#selector".
                If the '#selector' part is omitted, the entire content of the
                file is merged.
            keychain_str (str, optional): A keychain string specifying the target
                location for the merge within the current tree. If the string
                ends with a '/', any existing data at that location will be
                deleted before the new data is merged. Defaults to None, which
                merges at the root.
            relative (bool): This parameter is present but currently has no effect.

        Side Effects:
            The instance's `root_dir` attribute is updated to the directory of the
            file that was merged.

        Returns:
            None

        Examples:
            Assuming `external.yaml` contains:
            ```yaml
            section_a:
              key1: value1
            section_b:
              key2: value2
            ```

            And the current YAMLator instance `yt` contains `{'existing_data': ...}`.
            To merge `section_a` from `external.yaml` into `existing_data/`,
            replacing its contents:
            >>> yt.merge('external.yaml#section_a', 'existing_data/')
        """
        # stringify as the argument might be a pathlib.Path
        yaml_path_selector = str(yaml_path_selector)
        if not '#' in yaml_path_selector:
            # assume we want to load the whole yaml file
            yaml_path_selector += '#'

        _yaml_path,_selector = yaml_path_selector.split('#')
        if _DEBUG.YAMLator:
            ic(_yaml_path)
            ic(_selector)

        if keychain_str is None:
            _keychain = []
        else:
            _keychain = keychain_str.split('/')
        if _keychain and _keychain[-1] == '':
            _keychain.pop()
            self.pop(_keychain)

        _yaml_path = pathlib.Path(_yaml_path)
        _root_dir = _yaml_path.parent

        _yt = YAMLTransformer(
            self.__class__(
                OrderedDict(
                    _merge_data=f'./{_yaml_path.name}#{_selector}'
                ),
                root_dir=_root_dir
            )
        )

        # notice: no transforms are done on the file to be merged
        # but YAMLTransformer is recursively called
        _yt.evaluate()
        self.overlay(_yt.get('/_merge_data/'),_keychain)
        # merging changes the root_dir...is that ok?
        self.root_dir = _root_dir

    def set_config_attrs(self,set_all=False):
        """Populates the instance with attributes from the configuration data.

        This method provides convenient, attribute-style access to configuration
        data by populating the instance with uppercase attributes that correspond
        to keys in the YAML tree. Attribute values are automatically converted
        to their appropriate Python types based on the object database's rules.

        The method traverses the tree depth-first. If the same key name
        (e.g., 'HOST') appears at multiple levels of nesting, the value from
        the deepest path will be the one that is set as the attribute.

        Args:
            set_all (bool, optional): Acts as a schema validator.
                - If False (the default), the method will only set attributes
                  that have been pre-defined on the class.
                - If True, the method will create and set an attribute for
                  every key found in the YAML data, regardless of whether it
                  was pre-defined.

        Returns:
            None: The method modifies the instance in-place.
        """
        _is_already_set = dict()

        def _depth_first(_node,_keychain):
            if _DEBUG.YAMLator:
                ic(_keychain)
            _node_keys = copy(list(_node.keys()))
            _node_keys.reverse()
            for _node_key in _node_keys:
                _attr_name = _node_key.replace('-', '_').upper()
                if _DEBUG.YAMLator:
                    ic(_attr_name)
                # strip the leading underscore
                if _attr_name.startswith('_'):
                    _attr_name = _attr_name[1:]

                if hasattr(self,_attr_name) or set_all:
                    if _is_already_set.get(_attr_name,False):
                        if _DEBUG.CONFIG_ATTRS:
                            _msg = f'{_attr_name} already set'
                            ic(_msg)
                        continue
                    try:
                        if isinstance(getattr(type(self), _attr_name), property):
                            # ignore any uppercase attribute that is a property
                            continue
                    except AttributeError:
                        pass

                    # -----------------------------------------------------------
                    # this can cause recursion, because we override get() to call super().get()
                    # only ever set the deepest value of a given key; get() is dfs
                    # _value = self.get(_node_key)
                    # -----------------------------------------------------------
                    _value = _node.get(_node_key)

                    if _DEBUG.CONFIG_ATTRS:
                        ic(_node_key)
                        ic(_value)

                    if isinstance(_value, str) and is_variable_token(_value):
                        # ignore unsubbed variables
                        continue

                    if isinstance(_value,OrderedDict):
                        _value = Tree(_value)

                    # # TODO: this definition of True needs to be formalized for other transformers, like IfKey
                    # if self.objdb.key_type(_node_key) == 'bool' and isinstance(_value,Tree):
                    #     if any(map(lambda x: x.startswith('))?'), _node.keys())):
                    #         return None
                    #     else:
                    #         # a boolean key with a tree value (that is not an if key node) is True (?)
                    #         return True

                    if isinstance(_value,list):
                        _object_value = list(map(lambda x:self.get_object(_node_key,x),_value))
                    else:
                        _object_value = self.get_object(_node_key, _value)

                    if _DEBUG.CONFIG_ATTRS:
                        ic(_object_value)
                    setattr(self, _attr_name, _object_value)
                    _is_already_set[_attr_name] = True

        # we set objects on demand with a special object getter
        self.visit(post_process=_depth_first)

    def get_config_attrs(self):
        """Retrieves a summary of the instance's configuration attributes.

        This method is an introspection tool that scans the YAMLator instance
        for all public, uppercase attributes, which are typically set by a
        prior call to `set_config_attrs`.

        Returns:
            dict: A dictionary where keys are the uppercase attribute names
                (e.g., 'SERVER_HOST') and values are the string
                representations of the corresponding attribute values.

        Examples:
            >>> yaml_content = "SERVER_HOST: 'localhost'\\nSERVER_PORT: 8080"
            >>> yt = YAMLator(yaml_content)
            >>> yt.set_config_attrs(set_all=True)
            >>> yt.get_config_attrs()
            {'SERVER_HOST': 'localhost', 'SERVER_PORT': '8080'}

        """
        _data = {}
        for _config_datum in list(filter(lambda x:x == x.upper() and not x.startswith('_'),self.__dir__())):
            _data.update({_config_datum:str(getattr(self,_config_datum))})
        return _data

    def validate(self, context_tree=None) -> list:
        """Runs all available pre-flight validation checks.

        This method orchestrates the validation process by calling the `validate()`
        method on its registered transformer utilities. It is designed to support
        a gradual rollout of the validation framework, safely ignoring any
        utilities that have not yet implemented the `validate()` method.

        Args:
            context_tree (YAMLator, optional): An external context tree to be
                used for cross-file validation. Defaults to None.

        Returns:
            list: A consolidated list of all warning and error strings found
                by the validators.
        """
        all_issues = []
        for utility_name in DEFAULT_UTILITIES.keys():
            utility = getattr(self, utility_name)(context_tree=context_tree)
            if hasattr(utility, 'validate'):
                if _DEBUG.VALIDATE:
                    _msg = f"running {utility_name} validation"
                    ic(_msg)
                issues = utility.validate()
                all_issues.extend(issues)
        return all_issues

    def check_subs(self):
        _unsubbed = Tree()

        def _assert_not_unsubbed(value,keychain):
            if not isinstance(value,list):
                _values = [value]
            else:
                _values = value
            for _value in _values:
                ic(_value)
                if type(_value) is str:
                    _tokens = ValueTransformer()._tokenize(_value)
                    if any(list(map(lambda x:is_variable_token(x),_tokens))):
                        _unsubbed.get(keychain,value)

        self.visit(value_process=_assert_not_unsubbed)
        return _unsubbed



from ..constants import *
from ..tree import *
from .KeyChainTransformer import *

from . import DEBUG

class ValueTransformerException(Exception):
    pass

class ValueTransformer(KeyChainTransformer):
    name = 'transform_values'
    def _do_not_evaluate(self, value, keychain):
        if keychain and keychain[-1].startswith('_'):
            if DEBUG.ValueTransformer:
                _msg = f'NOT EVALUATING {keychain}'
                ic(_msg)
            return True
        return False

    def _value_evaluate(self, value, keychain):
        '''generically transform in-place a multiline value using self._transform()'''
        # if DEBUG.ValueTransformer:
        #     ic()
            # ic(keychain)

        if self._do_not_evaluate(value, keychain):
            return

        if not isinstance(value, list):
            _tmp_values = [str(value)]
        else:
            _tmp_values = list(map(str, value))

        # if DEBUG.ValueTransformer:
        #     ic(_tmp_values)

        # the list of transformed values
        _transformed_values = []

        _transformed_lines = []
        # are the results of _transform() Tree values?
        _are_tree_values = False

        for _i, _tmp_value in enumerate(_tmp_values):
            _tokenized_lines = list(map(lambda x: self._tokenize(x), _tmp_value.split('\n')))

            # if DEBUG.ValueTransformer:
            #     ic(_tmp_value)
            #     ic(_tokenized_lines)

            _transformed_lines = []
            for _tokenized_line in _tokenized_lines:
                _transformed_line = []
                for _token in _tokenized_line:

                    # if DEBUG.ValueTransformer:
                    #     ic(_token)

                    _transformed_token = _token
                    # ------------------------------
                    _extract = self._extract(_token)
                    # ------------------------------
                    if DEBUG.ValueTransformer:
                        ic(_extract.groups()) if _extract else None
                    # MUST reset this!
                    _are_tree_values = False
                    if _extract is not None:
                        _transformed_token = self._transform(_extract.groups(), keychain)

                        if DEBUG.ValueTransformer:
                            ic(_transformed_token)

                    if _transformed_token is None:
                        _transformed_token = _token
                    elif isinstance(_transformed_token, Tree):
                        # we need to recreate Tree objects from str reps when we self.get() below on transformed values
                        # this works but is awkward
                        _are_tree_values = True
                    elif isinstance(_transformed_token, list):
                        _transformed_token = ' '.join(map(str, _transformed_token))
                    elif isinstance(_transformed_token, str) and re.match(
                            # TODO: this needs refinement
                            # DO NOT SUB IN ANY UNRESOLVED SELECTORS
                            # unevaluated yaml path selector strings NEVER get subbed as values
                            # f"(.*?)({PathValueTransformer.posix_abolute_regex}|{PathValueTransformer.posix_relative_regex}#.*)(.*)",
                            # f".*?({REGEXES.POSIX_RELATIVE}|{REGEXES.POSIX_ABSOLUTE})#.*",
                            rf"^[^#](.*?#.*)",
                            _transformed_token):
                        if DEBUG.ValueTransformer:
                            _msg = 'throwing away transformed token'
                            ic(_msg)
                            ic(_transformed_token)
                        _transformed_token = _token
                    # make sure we append strings
                    _transformed_line.append(str(_transformed_token))
                _transformed_lines.append(''.join(_transformed_line))
            for _line in _transformed_lines:
                _transformed_values.append(_line)

        # if DEBUG.ValueTransformer:
        #     ic(_transformed_values)

        if _are_tree_values:
            # now insert Tree objects and return
            if DEBUG.ValueTransformer:
                ic(_are_tree_values)
                ic(_transformed_values)

            try:
                # note how this handles lists of trees elegantly
                # TODO: but it overwrites existing trees and does not overlay them!
                _t = Tree()
                _trees = list(map(lambda x: _t.load(io.StringIO(x)), _transformed_values))
                if DEBUG.ValueTransformer:
                    ic(_trees)
                    ic(keychain)
                # list(map(lambda x: self.get([''] + keychain, x), _trees))
                list(map(lambda x: self.overlay(x,[''] + keychain), _trees))
                return
            except TreeCreationException as e:
                if DEBUG.ValueTransformer:
                    _msg = 'failed to _transform() tree'
                    ic(_msg)
                return
            except Exception as e:
                if DEBUG.ValueTransformer:
                    ic(e)

        if isinstance(value, list):
            value = _transformed_values
        else:
            value = '\n'.join(_transformed_values)

        self.get([''] + keychain, value)



from ..constants import *
from ..tree import *
from .PathValueTransformer import *


class PythonTransformerException(Exception):
    pass


class PythonTransformer(PathValueTransformer):
    ext = 'py'

    def __init__(self, odict_or_tree):
        if DEBUG.PythonTransformer: ic()
        super(PythonTransformer,self).__init__(odict_or_tree)

    def _loader(self,stream):
        # see https://github.com/berkerpeksag/astor/blob/master/astor/file_util.py
        fstr = stream.read()
        fstr = fstr.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
        if not fstr.endswith(b'\n'):
            fstr += b'\n'
        return ast.parse(fstr, filename=stream.name)

    def _translate_keychain(self, _py_keychain):
        '''keychains in python are strings of the form x__y__z -> ))(x/y/z) e.g.'''
        if '__' in _py_keychain:
            # keychain type
            return f"{KEY_OR_KEYCHAIN_OP}{KEYCHAIN_LEFT_BOUND}{_py_keychain.replace('__', KEYCHAIN_SEP)}{KEYCHAIN_RIGHT_BOUND}"
        else:
            # key type
            return f"{KEY_OR_KEYCHAIN_OP}{_py_keychain}"

    def select(self, stream, _py_func_name):
        # holds the lines of a multiline Tree value
        _exec_node_lines = []

        class FuncVisitor(ast.NodeVisitor):
            def _bind_arguments(_self, arguments, line):
                for _argument in map(lambda x: astor.to_source(x).strip(), arguments):
                    # translate the python keychain syntax into substitution syntax
                    _new_argument = self._translate_keychain(_argument)
                    ############################################
                    # TODO: - and _ are ambiguous in key names for legacy reasons! Dumb!
                    ############################################
                    # print(f"!! altering referred key name from {_new_argument} to {_new_argument.replace('_','-')}")
                    _new_argument = _new_argument.replace('_','-')
                    # this is really dumb. There must be a better way
                    line = line.replace(f'{_argument}', f"'{_new_argument}'")
                return line

            def visit_Import(_self, node):
                _exec_node_lines.append(astor.to_source(node))

            def visit_ImportFrom(_self, node):
                _exec_node_lines.append(astor.to_source(node))

            def visit_FunctionDef(_self, node):
                # alter the body of the functions to use SubstitutionTransformer syntax
                if node.name == _py_func_name:
                    [_exec_node_lines.append(_self._bind_arguments(node.args.args, x)) for x in
                                                             list(map(astor.to_source, node.body))]


        FuncVisitor().visit(self._loader(stream))

        return Tree(OrderedDict(python=''.join(_exec_node_lines)))



"""
Deprecated, tree values are always string reps of objects, upper case vars are set with object values
"""

from ..evaluators import AbstractEvaluator
from ..transformers import Transformer, TransformerException, AtTransformer
from ..constants import *

class ObjectValueEvaluatorException(Exception):
    pass

class ObjectValueEvaluator(AbstractEvaluator):

    def _all_tokens(self,value_str):
        return
    @classmethod
    def _object_map(cls, key, values):
        # TODO we need a standard to decode obj types from key names
        _new_values = []
        for _value in values:
            # note: special variables don't tokenize like regular variables!!
            _tokens = Transformer()._tokenize(_value) + AtTransformer()._tokenize(_value)
            if any(list(map(lambda x:is_variable_token(x),_tokens))):
                _new_values.append(_value)
                continue
            if '-dir' in key or '-path' in key:
                _value = pathlib.Path(_value).expanduser()
                if '-dir' in key:
                    # directory creation
                    if not _value.exists():
                        try:
                            _value.mkdir(parents=True, exist_ok=True)
                        except PermissionError:
                            raise ObjectValueEvaluatorException
                            # print(f"permision denied to create {_value}")
                else:
                    # path creation
                    if not _value.exists():
                        try:
                            _value.touch(exist_ok=True)
                        except PermissionError:
                            print(f"permission denied to create {_value}. This could be fatal.")
            elif '-url' in key:
                # TODO test url
                _value = urlpath.URL(_value)
            # elif 'use-' in key or 'is-' in key or 'has-' in key:
            elif key.startswith('use-') or key.startswith('is-') or key.startswith('has-'):
                if isinstance(_value,str):
                    if not re.match(REGEXES.BOOL_REGEX,_value):
                        raise TransformerException(f"{_value} is not a recognized boolean string")
                    if _value in TRUE_VALUES:
                        _value = True
                    elif _value in FALSE_VALUES:
                        _value = False
                    else:
                        raise TransformerException("Regex failure")
                elif isinstance(_value,bool):
                    pass
                else:
                    raise TransformerException(f"Wrong value type: {type(_value)} for key: {key}")
            _new_values.append(_value)
        return _new_values

    def _value_evaluate(self, value, keychain):
        _key = keychain[-1]
        _value = value
        if not isinstance(_value, list):
                _tmp_values = [str(_value)]
        else:
            _tmp_values = list(map(str,_value))

        _new_values = self._object_map(_key,_tmp_values)

        if isinstance(value,list):
            value = _new_values
        else:
            value = _new_values.pop()
        self.get(keychain, value)



from __future__ import annotations

from ..constants import *
from ..tree import *
from .AbstractValidator import  *
from .issues import *

class _DEBUG:
    ValueValidator = True


class ValueValidator(AbstractValidator):
    """
    Validates a tree for issues related to variable substitution, such as
    circular dependencies, and undefined or unused variables, before the
    main transformation process is run.
    """
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

    def validate(self, context_tree: AbstractValidator = None) -> list[ValidationResult]:
        """Runs all value-related validation checks.

        Args:
            context_tree (AbstractValidator, optional): An external context tree
                for resolving variables. Defaults to None.

        Returns:
            list[ValidationResult]: A list of all validation issues found.
        """
        issues = []
        issues.extend(self._find_circular_dependencies())
        issues.extend(self._find_undefined_variables(context_tree))
        return issues


    def _find_circular_dependencies(self) -> list[ValidationResult]:
        """Detects cyclical dependencies in the variable graph.

        Returns:
            list[ValidationResult]: A list of validation results for any cycles found.
        """
        _reduced_tree = self.reduce()
        _visited = []
        _issues = []

        def _find_cycles(node, keychain):
            keychain_str = ''
            # Parse the variable token to get the keychain string to test
            for var_token in node:
                if var_token.startswith(KEY_OR_KEYCHAIN_OP):
                    if '{' in var_token and '}' in var_token:
                        # Handles cases like ')){a/b/c}'
                        start_index = len(KEY_OR_KEYCHAIN_OP) + 1
                        keychain_str = var_token[start_index:-1]
                    else:
                        # Handles cases like '))a-key'
                        start_index = len(KEY_OR_KEYCHAIN_OP)
                        keychain_str = var_token[start_index:]

                    if keychain_str in _visited:
                        _issues.append(ValidationResult(
                            issue_type=ValidationIssue.CIRCULAR_DEPENDENCY,
                            message=ValidationIssue.CIRCULAR_DEPENDENCY.value.format(path=' -> '.join(_visited))
                        ))
                        while _visited:
                            _visited.pop()
                        continue

                    _visited.append(keychain_str)
                    entry_keychain = keychain_str.split('/')
                    raise TreeVisitRestartException(entry_keychain)

        _reduced_tree.visit(value_process=_find_cycles)

        return _issues

    def _find_undefined_variables(self, context_tree: AbstractValidator = None) -> list[ValidationResult]:
        """Finds variables whose referenced keychains cannot be resolved in the tree.

        This method works by parsing each used variable token (e.g., ')){my/key}')
        to extract its keychain ('my/key'). It then attempts to resolve that
        keychain using `self.get()`. If the keychain cannot be resolved in either
        the current tree or the optional context_tree, it is marked as undefined.

        Args:
            context_tree (AbstractValidator, optional): An external tree to use
                as a secondary source for variable definitions. Defaults to None.

        Returns:
            list[ValidationResult]: A list of issues for any undefined variables.
        """
        issues = []
        used_vars_map = self.invert()

        for var_token in used_vars_map.keys():
            # Parse the variable token to get the keychain string to test
            keychain_str = ''
            if '{' in var_token and '}' in var_token:
                # Handles cases like ')){a/b/c}'
                start_index = len(KEY_OR_KEYCHAIN_OP) + 1
                keychain_str = var_token[start_index:-1]
            else:
                # Handles cases like '))a-key'
                start_index = len(KEY_OR_KEYCHAIN_OP)
                keychain_str = var_token[start_index:]

            # An empty keychain (e.g., from '))/' or ')){/}') refers to the
            # root, which always exists, so we can skip checking it.
            if not keychain_str or keychain_str == '/':
                continue

            # Check if the keychain is defined in self or the context_tree
            is_defined = False
            try:
                # Check the current tree first
                self.get(keychain_str)
                is_defined = True
            except KeyError:
                # If not in the current tree, check the context_tree if it exists
                if context_tree:
                    try:
                        context_tree.get(keychain_str)
                        is_defined = True
                    except KeyError:
                        # Not defined in the context_tree either
                        pass

            # If the keychain was not resolved in any context, it's undefined
            if not is_defined:
                issues.append(ValidationResult(
                    issue_type=ValidationIssue.UNDEFINED_VARIABLE,
                    message=ValidationIssue.UNDEFINED_VARIABLE.value.format(variable=var_token)
                ))

        return issues


from ..constants import *
from ..tree import *
from .ValueTransformer import *

from . import DEBUG

class IfTransformerException(Exception):
    pass


# TODO: using self.context needs tests
class IfTransformer(ValueTransformer):
    name = 'transform_ifs'
    match_regex = rf"{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.IF}"
    extract_regex = rf'{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.IF_EXP}'

    def __init__(self, odict_or_tree=None, context_tree=None, **kwargs):
        odict_or_tree = OrderedDict() if odict_or_tree is None else odict_or_tree
        # super(IfTransformer, self).__init__(odict_or_tree)
        super().__init__(odict_or_tree)
        self.context = context_tree

    def reduce_logical_exp(self, logical_sub_exps):
        _reduced_logical_exps = []
        for _logical_sub_exp in logical_sub_exps:
            _logical_connective = None
            _not_if = False
            if _logical_sub_exp.strip()[0] == REGEXES.OR_SEP:
                _logical_connective = REGEXES.OR_SEP
                _logical_sub_exp = _logical_sub_exp.strip()[1:]
            elif _logical_sub_exp.strip()[0] == REGEXES.AND_SEP:
                _logical_connective = REGEXES.AND_SEP
                _logical_sub_exp = _logical_sub_exp.strip()[1:]
            else:
                _logical_sub_exp = _logical_sub_exp.strip()
            _logical_terms = _logical_sub_exp.split('=')
            if len(_logical_terms) == 2:
                assert _logical_terms[0][-1] == '!'
                _not_if = True
                _left_key = _logical_terms[0][:-1].strip()
                _right_key = _logical_terms[1].strip()
            elif len(_logical_terms) == 3:
                assert _logical_terms[1] == ''
                _left_key = _logical_terms[0].strip()
                _right_key = _logical_terms[2].strip()
            else:
                # I think we have a boolean-key
                if DEBUG.IfTransformer:
                    _msg = f'testing {_logical_sub_exp} as a boolean-key'
                    ic(_msg)

                _negate_test = False
                if '!' in _logical_sub_exp:
                    _logical_sub_exp = _logical_sub_exp[_logical_sub_exp.index('!') + 1:].strip()
                    _negate_test = True
                if DEBUG.IfTransformer:
                    ic(_negate_test)
                # a key does not exist yet
                try:
                    if self.context is not None:
                        _test_result = self.context.get(_logical_sub_exp.strip())
                    else:
                        _test_result = self.get(_logical_sub_exp.strip())

                except KeyError:
                    if DEBUG.IfTransformer:
                        _msg = f'get({_logical_sub_exp.strip()} not found!)'
                    return None
                if DEBUG.IfTransformer:
                    ic(_test_result)

                if isinstance(_test_result, str):
                    if not re.match(REGEXES.BOOL_VALUE, _test_result):
                        raise IfTransformerException(f"{_test_result} is not a recognized boolean string")
                    if _test_result in TRUE_VALUES:
                        _test_result = True if not _negate_test else False
                    elif _test_result in FALSE_VALUES:
                        _test_result = False if not _negate_test else True
                # elif isinstance(_test_result, Tree):
                #     # TODO: this definition of True needs to be formalized for other transformers, like IfKey
                #     if any(map(lambda x: x.startswith('))?'), _test_result.keys())):
                #         return None
                #     else:
                #         # a boolean key with a tree value (that is not an if key node) is True (?)
                #         _test_result = True if not _negate_test else False
                elif not isinstance(_test_result, bool):
                    return None

                _reduced_logical_exps.append((_test_result,_logical_connective))
                continue

            _slice = None
            _local_context_keychain = None
            _m = re.match(rf"^({REGEXES.KEYCHAIN}{REGEXES.KEY})\[({REGEXES.DATA_INDEX})\]$",_left_key)
            if _m:
                _left_key,_data_index = _m.groups()
                if re.match(REGEXES.SLICE,_data_index):
                    _slice = _data_index
                else:
                    _local_context_keychain = _data_index

            if DEBUG.IfTransformer:
                ic(_logical_terms)
                ic(_left_key)
                ic(_slice)
                ic(_local_context_keychain)
                ic(_right_key)
                ic(_not_if)
                ic(_logical_connective)

            _if = not _not_if
            if _left_key[0] == _left_key[-1] == "'":
                # strip the quotes on strings
                _left_value = _left_key[1:-1]
            else:
                # it's a key (unless it doesn't exist; error is caught by _transform())
                if self.context is not None:
                    _left_value = self.context.get(_left_key)
                else:
                    _left_value = self.get(_left_key)
            if _slice:
                assert isinstance(_left_value,str) or isinstance(_left_value,list)
                _left_value = eval(f'_left_value[{_slice}]')
            elif _local_context_keychain:
                assert isinstance(_left_value,Tree)
                _left_value = _left_value.get(_local_context_keychain)

            if _right_key[0] == _right_key[-1] == "'":
                _right_value = _right_key[1:-1]
            else:
                # it's a key (unless it doesn't exist; error is caught by _transform())
                if self.context is not None:
                    _right_value = self.context.get(_right_key)
                else:
                    _right_value = self.get(_right_key)

            if DEBUG.IfTransformer:
                ic(_left_value)
                ic(_right_value)

            if type(_left_value) != type(_right_value):
                if DEBUG.IfTransformer:
                    _msg = 'Cannot compare values'
                    ic(_msg)
                return None

            _reduced_logical_exps.append((
                _left_value == _right_value if _if else _left_value != _right_value,_logical_connective))

        if DEBUG.IfTransformer:
            ic(_reduced_logical_exps)

        assert _reduced_logical_exps[0][-1] is None
        # the rest should have a connective
        assert all([_rle[-1] is not None for _rle in _reduced_logical_exps[1:]])

        _current_logical_state = None
        for _reduced_logical_exp in _reduced_logical_exps:
            _equality_truth, _logical_connnective = _reduced_logical_exp
            if _current_logical_state is None:
                _current_logical_state = _equality_truth
            elif _logical_connnective == REGEXES.OR_SEP:
                _current_logical_state = _current_logical_state or _equality_truth
            elif _logical_connnective == REGEXES.AND_SEP:
                _current_logical_state = _current_logical_state and _equality_truth
            else:
                raise IfTransformerException(f'unknown logical connective: {_logical_connnective}')

            if DEBUG.IfTransformer:
                ic(_current_logical_state)

        return _current_logical_state

    def _transform(self, parameters, keychain):
        _logical_exp,_key_if_true,_key_if_false = parameters
        _logical_exp = _logical_exp.strip()
        _key_if_true = _key_if_true.strip()
        _key_if_false = _key_if_false.strip() if _key_if_false is not None else _key_if_false
        _logical_sub_exps = re.findall(REGEXES.LOGICAL_EXP,_logical_exp)
        if DEBUG.IfTransformer:
            ic()
            ic(_logical_exp)
            ic(_key_if_true)
            ic(_key_if_false)
            ic(_logical_sub_exps)
        # TODO: review this error handling
        try:
            _test_result = self.reduce_logical_exp(_logical_sub_exps)
        except KeyError:
            if DEBUG.IfTransformer:
                _msg = f'self.reduce_logical_exp({_logical_sub_exps}) raised KeyError'
                ic(_msg)
            # we're missing a key in the tree
            return None

        if DEBUG.IfTransformer:
            ic(_test_result)

        if _test_result is None:
            return None

        if _test_result is True:
            if DEBUG.IfTransformer:
                ic(_key_if_true)
            try:
                _return_value = self.get(_key_if_true)
            except KeyError:
                if _key_if_true[0] == _key_if_true[-1] == "'":
                    _return_value = _key_if_true[1:-1]
                else:
                    _return_value =  _key_if_true
        else:
            if DEBUG.IfTransformer:
                ic(_key_if_false)
            try:
                _return_value = self.get(_key_if_false) if _key_if_false is not None else ''
            except KeyError:
                if _key_if_false[0] == _key_if_false[-1] == "'":
                    _return_value =  _key_if_false[1:-1]
                else:
                    _return_value =  _key_if_false

        if DEBUG.IfTransformer:
            ic(_return_value)

        return _return_value


class IfTransformerUtility(IfTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.evaluate()

class IfKeyTransformer(IfTransformer):
    name = 'transform_if_keys'
    extract_regex = rf'{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.IF_KEY_EXP}'

    def __init__(self, odict_or_tree=None, context_tree=None, **kwargs):
        # super(IfKeyTransformer,self).__init__(odict_or_tree)
        super().__init__(odict_or_tree)
        self.context = context_tree
        # if DEBUG.IfKeyTransformer:
        #     ic()

    def _replace_node_key(self, node, node_key, new_node_key, new_key_value):

        if DEBUG.IfKeyTransformer:
            ic(node_key)
            ic(new_node_key)
            ic(new_key_value)

        # short circuit; is this still necessary?
        # Yes, for replacing an if-key with a str value !!!
        #  NO! you would need to set node = new_key_value, which does not work
        # if new_node_key is None and not isinstance(new_key_value,OrderedDict):
        #     node[node_key] = new_key_value
        #     return new_key_value

        _node_copy = deepcopy(node)
        _key_index = list(_node_copy.keys()).index(node_key)

        for _key in list(_node_copy.keys())[:_key_index]:
            if DEBUG.IfKeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

        if DEBUG.IfKeyTransformer:
            _msg = f'replacing {node_key} with {new_key_value}'
            ic(_msg)

        _ = node.pop(node_key)
        if not isinstance(new_key_value,OrderedDict):
            node[new_node_key] = new_key_value
        else:
            for _key in new_key_value.keys():
                node[_key] = new_key_value.get(_key)

        for _key in list(_node_copy.keys())[_key_index + 1:]:
            if DEBUG.IfKeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

    def _value_evaluate(self, value, keychain):
        pass

    def _pre_evaluate(self,node,keychain):

        if any(map(lambda x: x.startswith('))?'), node.keys())):
            if DEBUG.IfKeyTransformer:
                ic()
                ic(self.__class__)
                ic(keychain)
                ic(node.keys())
        else:
            return None

        # TODO: forward slashes and string string values are problematic
        for _node_key in copy(list(node.keys())):
            if not _node_key.startswith('))?'):
                continue

            if DEBUG.IfKeyTransformer:
                ic(_node_key)

            _transformed_lines = []
            _transformed_key = []
            _if = True

            _token = _node_key

            _transformed_token = _token
            _parameters = self._extract(_token)

            _logical_exp,_forward_slash = _parameters.groups()
            _logical_sub_exps = re.findall(REGEXES.LOGICAL_KEY_EXP, _logical_exp)

            _node_copy = deepcopy(node)
            _key_index = list(_node_copy.keys()).index(_node_key)
            _branches = node.get(_node_key)
            if isinstance(_branches,OrderedDict):
                assert len(list(_branches.keys())) < 3
                if not _forward_slash:
                    _true_branch = OrderedDict({list(_branches.keys())[0]: list(_branches.values())[0]})
                else:
                    _true_branch = list(_branches.values())[0]
                if len(list(_branches.keys())) == 2:
                    if not _forward_slash:
                        _false_branch = OrderedDict({list(_branches.keys())[1]:list(_branches.values())[1]})
                    else:
                        _false_branch = list(_branches.values())[1]
                else:
                    _false_branch = None
            else:
                _true_branch = _branches
                _false_branch = None

            if DEBUG.IfKeyTransformer:
                ic(_branches)
                ic(_key_index)
                ic(_forward_slash)
                ic(_true_branch)
                ic(_false_branch)
                ic(_logical_exp)
                ic(_logical_sub_exps)
                # if self.context:
                #     ic(self.context.print())
                # else:
                #     ic(self.context)

            # TODO: review this error handling
            try:
                _test_result = self.reduce_logical_exp(_logical_sub_exps)
            except Exception as e:
                if DEBUG.IfKeyTransformer:
                    _msg = f'{_logical_sub_exps} failed. Returning None'
                    ic(_msg)
                    ic(e)
                return None

            if DEBUG.IfKeyTransformer:
                ic(_test_result)

            if _test_result is None:
                return None

            if _test_result:

                if DEBUG.IfKeyTransformer:
                    _msg = f'{_logical_exp} is TRUE'
                    ic(_msg)

                # are we key order preserving here? NO! fix it...
                if _if:
                    if not _forward_slash:
                        self._replace_node_key(node,_node_key,list(_branches.keys())[0],_true_branch)
                    else:
                        self._replace_node_key(node, _node_key, None, _true_branch)

                elif not _if:
                    if _false_branch:
                        if not _forward_slash:
                            self._replace_node_key(node, _node_key,list(_branches.keys())[1], _false_branch)
                        else:
                            self._replace_node_key(node, _node_key, None, _false_branch)

                    else:
                        _ = node.pop(_node_key)

            else:

                if DEBUG.IfKeyTransformer:
                    _msg = f'{_logical_exp} is FALSE'
                    ic(_msg)

                if not _if:
                    if not _forward_slash:
                        self._replace_node_key(node, _node_key,list(_branches.keys())[0], _true_branch)
                    else:
                        self._replace_node_key(node, _node_key, None, _true_branch)
                elif _if:
                    if _false_branch:
                        if not _forward_slash:
                            self._replace_node_key(node, _node_key,list(_branches.keys())[1], _false_branch)
                        else:
                            self._replace_node_key(node, _node_key, None, _false_branch)
                    else:
                        _ = node.pop(_node_key)


class IfKeyTransformerUtility(IfKeyTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from ..tree import *
from ..constants import *
from .ValueTransformer import ValueTransformer

from . import DEBUG

class BangTransformerException(Exception):
    pass


class BangTransformer(ValueTransformer):
    match_regex = rf"{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.BANG}"
    extract_regex = rf"{re.escape(KEY_OR_KEYCHAIN_OP)}{REGEXES.BANG_EXP}"

    name = 'transform_bangs'
    def __init__(self, odict_or_tree=None, **kwargs):
        odict_or_tree = OrderedDict() if odict_or_tree is None else odict_or_tree
        super(BangTransformer, self).__init__(odict_or_tree)
        # if DEBUG.BangTransformer:
        #     ic()

        if isinstance(odict_or_tree, Tree):
            self.driver_source = odict_or_tree
        else:
            self.driver_source = None

    def _transform(self, parameters, keychain):
        _method_call, = parameters
        if DEBUG.BangTransformer:
            ic(_method_call)

        class _MethodParser(ast.NodeVisitor):
            def visit_Call(self,node):
                self.method_name = astor.to_source(node.func).strip()
                self.arguments = list(map(lambda x:astor.to_source(x).strip(),node.args))
                self.keywords = list(map(lambda x:x.arg.strip(),node.keywords))
                self.keyword_values = list(map(lambda x:astor.to_source(x.value).strip(),node.keywords))

        _mp = _MethodParser()
        _mp.visit(ast.parse(_method_call))

        if DEBUG.BangTransformer:
            ic(_mp.method_name)
            ic(_mp.arguments)
            ic(_mp.keywords)
            ic(_mp.keyword_values)

        # this might be useful someday...
        # getattr(self,_mp.method_name).__call__(*_mp.arguments,**dict(zip(_mp.keywords,_mp.keyword_values)))

        _computed_value = None
        if self.driver_source:
            try:
                _computed_value = eval(f'self.driver_source.{_method_call}')
                if DEBUG.BangTransformer:
                    # _msg = f'computed {_method_call} as {_computed_value}'
                    _msg = f'method call success'
                    ic(_msg)
                    ic(self.driver_source)
            except AttributeError:
                if DEBUG.BangTransformer:
                    _msg = f"{_method_call} not found (or the method threw AttributeError)!"
                    ic(_msg)
                    ic(self.driver_source)
                    ic(dir(self.driver_source))
                # Eventually, it will be found!
                # raise AttributeError
                _computed_value = None
            except Exception as e:
                if DEBUG.BangTransformer:
                    _msg = f"{_method_call} failed with {e}!"
                    ic(_msg)
                    ic(self.driver_source)
                _computed_value = None


        else:
            raise BangTransformerException(f"no driver source configured for {_method_call}")

        return _computed_value

class BangTransformerUtility(BangTransformer):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)


from ..constants import *
from .KeyChainTransformer import *

from . import DEBUG


class KeyTransformerException(Exception):
    pass


class KeyTransformer(KeyChainTransformer):
    name = 'transform_keys'
    def __init__(self, odict_or_tree=None, context_tree=None, allow_tree_subs=False):
        super(KeyTransformer,self).__init__(odict_or_tree, context_tree, False)

    def _replace_node_key(self, node, node_key, _transformed_key):

        if DEBUG.KeyTransformer:
            ic(node_key)
            ic(_transformed_key)
        _node_copy = deepcopy(node)
        _key_index = list(_node_copy.keys()).index(node_key)

        for _key in list(_node_copy.keys())[:_key_index]:
            if DEBUG.KeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

        if DEBUG.KeyTransformer:
            _msg = f'replacing {node_key} with {_transformed_key}'
            ic(_msg)

        _value = node.pop(node_key)
        if DEBUG.KeyTransformer:
            ic(_value)

        node[_transformed_key] = _value

        for _key in list(_node_copy.keys())[_key_index + 1:]:
            if DEBUG.KeyTransformer:
                _msg = f'moving {_key}'
                ic(_msg)
            _value = node.pop(_key)
            node[_key] = _value

    def _pre_evaluate(self, node, keychain):
        if DEBUG.KeyTransformer:
            ic()
            ic(keychain)
            ic(node.keys())

        for _node_key in copy(list(node.keys())):
            _tmp_key = _node_key

            _tokenized_key = self._tokenize(_tmp_key)
            if DEBUG.KeyTransformer:
                ic(_tokenized_key)

            _transformed_lines = []
            _transformed_key = []

            for _token in _tokenized_key:
                _transformed_token = _token
                _match = self._extract(_token)
                if _match is not None:
                    _transformed_token = self._transform(_match.groups(), keychain)

                if _transformed_token is None:
                    _transformed_token = _token
                elif isinstance(_transformed_token, list):
                    raise KeyTransformerException(f'Are you really trying to turn a list into a key?')
                    # _transformed_token = self._yaml_list_transformer(_transformed_token)
                if DEBUG.KeyTransformer:
                    ic(_transformed_token)

                # make sure we append strings
                _transformed_key.append(str(_transformed_token))

            _transformed_key = ''.join(_transformed_key)

            if _transformed_key != _node_key:
                self._replace_node_key(node,_node_key,_transformed_key)


class KeyTransformerUtility(KeyTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)